import sys
import wx
from wx.lib.agw import floatspin as fspin

is_wxPhoenix = 'phoenix' in wx.PlatformInfo

from wxutils.icons import get_icon

def FloatSpin(parent, value=0, action=None, tooltip=None,
                 size=(100, -1), digits=1, increment=1, **kws):
    """FloatSpin with action and tooltip"""
    if value is None:
        value = 0
    fs = fspin.FloatSpin(parent, -1, size=size, value=value,
                         digits=digits, increment=increment, **kws)
    if action is not None:
        fs.Bind(fspin.EVT_FLOATSPIN, action)
    if tooltip is not None:
        if is_wxPhoenix:
            fs.SetToolTip(tooltip)
        else:
            fs.SetToolTipString(tooltip)
    return fs

def FloatSpinWithPin(parent, value=0, pin_action=None,
                     tooltip='use last point selected from plot', **kws):
    """create a FloatSpin with Pin button with action"""
    fspin = FloatSpin(parent, value=value, **kws)
    bmbtn = wx.BitmapButton(parent, id=-1, bitmap=get_icon('pin'),
                            size=(25, 25))
    if pin_action is not None:
        parent.Bind(wx.EVT_BUTTON, pin_action, bmbtn)
    if is_wxPhoenix:
        bmbtn.SetToolTip(tooltip)
    else:
        bmbtn.SetToolTipString(tooltip)
    return fspin, bmbtn

class NumericCombo(wx.ComboBox):
    """
    Numeric Combo: ComboBox with numeric-only choices
    """
    def __init__(self, parent, choices, precision=3, fmt=None,
                 init=0, default_val=None, width=80):

        self.fmt = fmt
        if fmt is None:
            self.fmt = "%%.%if" % precision

        self.choices  = choices
        schoices = [self.fmt % i for i in self.choices]
        wx.ComboBox.__init__(self, parent, -1, '', (-1, -1), (width, -1),
                             schoices, wx.CB_DROPDOWN|wx.TE_PROCESS_ENTER)

        init = min(init, len(self.choices))
        if default_val is not None:
            if default_val in schoices:
                self.SetStringSelection(default_val)
            else:
                self.add_choice(default_val, select=True)
        else:
            self.SetStringSelection(schoices[init])
        self.Bind(wx.EVT_TEXT_ENTER, self.OnEnter)

    def OnEnter(self, event=None):
        self.add_choice(float(event.GetString()))

    def add_choice(self, val, select=True):
        if val not in self.choices:
            self.choices.append(val)
            self.choices.sort()
        self.choices.reverse()
        self.Clear()
        self.AppendItems([self.fmt % x for x in self.choices])
        if select:
            self.SetSelection(self.choices.index(val))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""This module provides base classes to implement models for
spectroscopy data.
"""
from __future__ import absolute_import, division
import os

from silx.gui import qt
import silx.io

from .items import (ExperimentItem,
                    GroupItem,
                    FileItem,
                    ScanItem)

from larch.utils.logging import getLogger
_logger = getLogger('larch.qtlib.view')


class HorizontalHeaderView(qt.QHeaderView):

    def __init__(self, parent=None):
        super(HorizontalHeaderView, self).__init__(qt.Qt.Horizontal, parent)

        # Some properties
        self.setStretchLastSection(True)
        # self.setSectionsMovable(True)

        # Context menu
        self.setContextMenuPolicy(qt.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.showContextMenu)

    def showContextMenu(self, position):
        menu = qt.QMenu('Horizontal Header View Menu', self)

        section = self.logicalIndexAt(position)

        action = qt.QAction('Add', self, triggered=self.append)
        menu.addAction(action)

        action = qt.QAction('Remove', self,
                            triggered=lambda: self.remove(section))
        menu.addAction(action)

        menu.exec_(self.mapToGlobal(position))

    def append(self):
        pass

    def remove(self, section):
        model = self.model()
        if not model.header[section].removable:
            _logger.info('The selected column cannot be removed')
            return
        model.setHeaderData(section, orientation=qt.Qt.Horizontal, value=None)
        view = self.parent()
        view.setItemsDelegates()


class TreeView(qt.QTreeView):

    def __init__(self, parent=None):
        super(TreeView, self).__init__(parent)

        # Header
        headerView = HorizontalHeaderView()
        self.setHeader(headerView)

        # Context menu
        self.setContextMenuPolicy(qt.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.showContextMenu)

        # Selection mode
        self.setSelectionMode(qt.QAbstractItemView.ExtendedSelection)

    def setModel(self, model):
        super(TreeView, self).setModel(model)
        self.setItemsDelegates()

    def setItemsDelegates(self):
        if self.model() is None:
            return

        header = self.model().header
        for i, _ in enumerate(header):
            delegate = header.delegate(i)
            if delegate is not None:
                self.setItemDelegateForColumn(i, delegate(parent=self))

    def showContextMenu(self, position):
        menu = qt.QMenu('Tree View Menu', self)

        action = qt.QAction(
            'Add Experiment', self, triggered=self.addExperiment)
        menu.addAction(action)

        # Get the index under the cursor.
        index = self.indexAt(position)
        item = self.model().itemFromIndex(index)

        if isinstance(item, ExperimentItem) or isinstance(item, GroupItem):
            action = qt.QAction('Add Group', self, triggered=self.addGroup)
            menu.addAction(action)

            action = qt.QAction('Load Files', self, triggered=self.loadFiles)
            menu.addAction(action)

        # If there are selected indexes, they can be removed or checked.
        if self.selectedIndexes():
            menu.addSeparator()
            action = qt.QAction(
                'Toggle Selected', self, triggered=self.toggleSelected)
            menu.addAction(action)
            action = qt.QAction(
                'Remove Selected', self, triggered=self.removeSelected)
            menu.addAction(action)

        if isinstance(item, ScanItem) and index.column() > 0:
            menu.addSeparator()
            action = qt.QAction(
                'Copy Value to Selected', self,
                triggered=lambda: self.copyValueToSelected(index))
            menu.addAction(action)

            action = qt.QAction(
                'Copy Value to Toggled', self,
                triggered=lambda: self.copyValueToToggled(index))
            menu.addAction(action)

        menu.exec_(self.mapToGlobal(position))

    def loadFiles(self):
        paths, _ = qt.QFileDialog.getOpenFileNames(
            self, 'Select Files to Load', os.getcwd(),
            'Data Files (*.spec *.hdf5);; All Files (*)')

        if not paths:
            return

        parent = self.selectionModel().selectedRows().pop()
        parentItem = self.model().itemFromIndex(parent)
        for path in paths:
            self.addFile(path, parentItem)

    def addExperiment(self, name=None):
        rootItem = self.model().rootItem
        row = rootItem.childCount()
        if name is None or not name:
            name = 'Experiment{}'.format(row)
        item = ExperimentItem(name=name, parentItem=rootItem)
        self.model().appendRow(item)
        return item

    def addGroup(self, name=None, parentItem=None):
        """Add a generic GroupItem at a given parentItem"""
        # Add the file to the last added item.
        if parentItem is None:
            parent = self.selectionModel().selectedRows().pop()
            parentItem = self.model().itemFromIndex(parent)
        row = parentItem.childCount()

        if name is None or not name:
            name = 'Group{}'.format(row)

        item = GroupItem(name, parentItem)
        self.model().appendRow(item)

    def addFile(self, path=None, parentItem=None):
        if path is None:
            return

        # Add the file to the last added experiment item.
        if parentItem is None:
            parentItem = self.model().rootItem.lastChild()

        try:
            data = silx.io.open(path)
        except OSError as e:
            _logger.warning(e)
            return

        # Create a tree item for the file and add it to the experiment item.
        name, _ = os.path.splitext(os.path.basename(path))
        item = FileItem(name, parentItem)
        self.model().appendRow(item)

        # Create a tree item for each scan. The parent item is now the
        # previous file item.
        # TODO: Make this more "intelligent" by using the command to
        # set better defaults for x, signal, etc.
        parentItem = item
        for scan in data:
            item = ScanItem(name=scan, parentItem=parentItem, data=data[scan])
            self.model().appendRow(item)

    def selectedItems(self):
        indexes = self.selectionModel().selectedRows()
        items = [self.model().itemFromIndex(index) for index in indexes]
        return items

    def scanItems(self):
        for index in self.model().visitModel():
            item = self.model().itemFromIndex(index)
            if isinstance(item, ScanItem):
                yield item

    def toggleSelected(self):
        for item in self.selectedItems():
            index = self.model().indexFromItem(item)
            try:
                if item.isChecked:
                    self.model().setData(
                        index, qt.Qt.Unchecked, qt.Qt.CheckStateRole)
                else:
                    self.model().setData(
                        index, qt.Qt.Checked, qt.Qt.CheckStateRole)
            except AttributeError:
                pass

    def removeSelected(self):
        items = self.selectedItems()
        parentItems = dict()
        for item in items:
            parentItem = item.parentItem
            remove = True
            while parentItem is not self.model().rootItem:
                if parentItem in items:
                    remove = False
                    break
                parentItem = parentItem.parentItem
            # If an ancestors is selected for removal, pass the item.
            if not remove:
                continue

            # Get the parent item for the current item.
            parentItem = item.parentItem
            if parentItem not in parentItems:
                parentItems[parentItem] = list()
            # Create a list with the positions of the children that are
            # going to be removed.
            parentItems[parentItem].append(item.childPosition())

        # Remove the rows from the parent.
        for parentItem in parentItems:
            rows = parentItems[parentItem]
            parent = self.model().indexFromItem(parentItem)
            for row in reversed(sorted(rows)):
                self.model().removeRow(row, parent)

    def copyValueToToggled(self, indexAt):
        """Copy the value under the cursor to the toggled items."""
        indexes = self.model().visitModel(columns=True)
        for index in indexes:
            item = self.model().itemFromIndex(index)
            if not item.isChecked or index == indexAt:
                continue
            elif index.column() == indexAt.column():
                value = self.model().data(indexAt)
                self.model().setData(index, value)

    def copyValueToSelected(self, indexAt):
        """Copy the value under the cursor to the selected indexes."""
        indexes = self.selectionModel().selectedIndexes()
        for index in indexes:
            if index == indexAt:
                continue
            elif index.column() == indexAt.column():
                value = self.model().data(indexAt)
                self.model().setData(index, value)



#!/usr/bin/env python
"""
Linear Combination panel
"""
import os
import time
import wx
import numpy as np

from functools import partial
from collections import OrderedDict

from larch.math import index_of

from larch.wxlib import (BitmapButton, FloatCtrl, FloatSpin, ToggleButton,
                         get_icon, SimpleText, pack, Button, HLine, Choice,
                         plotlabels, Check, CEN, RIGHT, LEFT)

from larch.xafs.xafsutils import etok, ktoe


from .xas_dialogs import EnergyUnitsDialog
from .taskpanel import TaskPanel

np.seterr(all='ignore')

# plot options:
mu_bkg  = '\u03bC(E) + \u03bc0(E)'
chie    = '\u03c7(E)'
chik    = '\u03c7(k)'
chikwin = '\u03c7(k) + Window(k)'
chirmag = '|\u03c7(R)|'
chirre  = 'Re[\u03c7(R)]'
chirmr  = '|\u03c7(R)| + Re[\u03c7(R)]'
noplot  = '<no plot>'

PlotOne_Choices = [mu_bkg, chie, chik, chikwin, chirmag, chirre, chirmr]
PlotAlt_Choices = [noplot] + PlotOne_Choices
PlotSel_Choices = [chie, chik, chirmag, chirre]


PlotCmds = {mu_bkg:  "plot_bkg({group:s}",
            chie:    "plot_chie({group:s}",
            chik:    "plot_chik({group:s}, show_window=False, kweight={plot_kweight:.0f}",
            chikwin: "plot_chik({group:s}, show_window=True, kweight={plot_kweight:.0f}",
            chirmag: "plot_chir({group:s}, show_mag=True, show_real=False",
            chirre:  "plot_chir({group:s}, show_mag=False, show_real=True",
            chirmr:  "plot_chir({group:s}, show_mag=True, show_real=True",
            noplot: None}

FTWINDOWS = ('Kaiser-Bessel', 'Hanning', 'Gaussian', 'Sine', 'Parzen', 'Welch')

CLAMPLIST = ('0', '1', '2', '5', '10', '20', '50', '100', '200', '500', '1000',
             '2000', '5000', '10000')

autobk_cmd = """autobk({group:s}, rbkg={rbkg: .3f}, e0={e0: .4f},
      kmin={bkg_kmin: .3f}, kmax={bkg_kmax: .3f}, kweight={bkg_kweight: .1f},
      clamp_lo={bkg_clamplo: .1f}, clamp_hi={bkg_clamphi: .1f})"""

xftf_cmd = """xftf({group:s}, kmin={fft_kmin: .3f}, kmax={fft_kmax: .3f},
      kweight={fft_kweight: .3f}, dk={fft_dk: .3f}, window='{fft_kwindow:s}')"""


defaults = dict(e0=-1.0, rbkg=1, bkg_kmin=0, bkg_kmax=None, bkg_clamplo=0,
                bkg_clamphi=1, bkg_kweight=1, fft_kmin=2.5, fft_kmax=None,
                fft_dk=4, fft_kweight=2, fft_kwindow='Kaiser-Bessel')

class EXAFSPanel(TaskPanel):
    """EXAFS Panel"""
    def __init__(self, parent, controller, **kws):
        TaskPanel.__init__(self, parent, controller,
                           configname='exafs_config',
                           config=defaults, **kws)
        self.skip_process = False
        self.last_plot = 'one'
        self.last_process_bkg = {}
        self.last_process_fft = {}

    def build_display(self):
        panel = self.panel
        wids = self.wids
        self.skip_process = True

        wids['plotone_op'] = Choice(panel, choices=PlotOne_Choices,
                                    action=self.onPlotOne, size=(175, -1))
        wids['plotalt_op'] = Choice(panel, choices=PlotAlt_Choices,
                                    action=self.onPlotOne, size=(175, -1))
        wids['plotsel_op'] = Choice(panel, choices=PlotSel_Choices,
                                    action=self.onPlotSel, size=(175, -1))

        wids['plotone_op'].SetStringSelection(chik)
        wids['plotsel_op'].SetStringSelection(chik)
        wids['plotalt_op'].SetStringSelection(noplot)

        plot_one = Button(panel, 'Plot This Group', size=(175, -1),
                          action=self.onPlotOne)

        plot_sel = Button(panel, 'Plot Selected Groups', size=(175, -1),
                          action=self.onPlotSel)


        saveconf = Button(panel, 'Save as Default Settings', size=(200, -1),
                          action=self.onSaveConfigBtn)

        def xxxFSWithPinPanel(name, value, **kws):
            s = wx.BoxSizer(wx.HORIZONTAL)
            self.wids[name] = FloatSpin(panel, value=value, **kws)
            bb = BitmapButton(panel, get_icon('pin'), size=(25, 25),
                              action=partial(self.onSelPoint, opt=name),
                              tooltip='use last point selected from plot')
            s.Add(self.wids[name])
            s.Add(bb)
            return s

        wids['plot_voffset'] = FloatSpin(panel, value=0, digits=2, increment=0.25,
                                         action=self.onProcess)
        wids['plot_kweight'] = FloatSpin(panel, value=2, digits=1, increment=1,
                                         action=self.onProcess, min_val=0, max_val=5)
        wids['plot_kweight_alt'] = FloatSpin(panel, value=2, digits=1, increment=1,
                                             action=self.onProcess,  min_val=0, max_val=5)

        opts = dict(digits=2, increment=0.1, min_val=0, action=self.onProcess)
        wids['e0'] = FloatSpin(panel, **opts)

        opts['max_val'] = 5
        wids['rbkg'] = FloatSpin(panel, value=1.0, **opts)

        opts['max_val'] = 125
        bkg_kmin = self.add_floatspin('bkg_kmin', value=0, with_pin=True, **opts)
        bkg_kmax = self.add_floatspin('bkg_kmax', value=20, with_pin=True, **opts)
        fft_kmin = self.add_floatspin('fft_kmin', value=0, with_pin=True, **opts)
        fft_kmax = self.add_floatspin('fft_kmax', value=20, with_pin=True, **opts)

        wids['fft_dk'] = FloatSpin(panel, value=3,  **opts)

        opts.update({'increment': 1, 'digits': 1, 'max_val': 5})
        wids['bkg_kweight'] = FloatSpin(panel, value=1, **opts)

        wids['fft_kweight'] = FloatSpin(panel, value=1, **opts)

        opts = dict(choices=CLAMPLIST, size=(80, -1), action=self.onProcess)
        wids['bkg_clamplo'] = Choice(panel, **opts)
        wids['bkg_clamphi'] = Choice(panel, **opts)

        wids['fft_kwindow'] = Choice(panel, choices=list(FTWINDOWS),
                                     action=self.onProcess, size=(150, -1))

        self.wids['is_frozen'] = Check(panel, default=False, label='Freeze Group',
                                       action=self.onFreezeGroup)

        def add_text(text, dcol=1, newrow=True):
            panel.Add(SimpleText(panel, text), dcol=dcol, newrow=newrow)

        def CopyBtn(name):
            return Button(panel, 'Copy', size=(60, -1),
                          action=partial(self.onCopyParam, name))


        panel.Add(SimpleText(panel, ' EXAFS Processing', **self.titleopts), dcol=5)

        panel.Add(plot_sel, newrow=True)
        panel.Add(self.wids['plotsel_op'], dcol=2)

        add_text('Vertical offset: ', newrow=False)
        panel.Add(wids['plot_voffset'], dcol=2)

        panel.Add(plot_one, newrow=True)
        panel.Add(self.wids['plotone_op'], dcol=2)

        add_text('Plot k weight: ', newrow=False)
        panel.Add(wids['plot_kweight'])

        add_text('Add Second Plot: ', newrow=True)
        panel.Add(self.wids['plotalt_op'], dcol=2)
        add_text('Plot2 k weight: ', newrow=False)
        panel.Add(wids['plot_kweight_alt'])


        panel.Add(HLine(panel, size=(500, 3)), dcol=6, newrow=True)

        panel.Add(SimpleText(panel, ' Background subtraction', size=(180, -1),
                             **self.titleopts), dcol=2, style=LEFT, newrow=True)
        panel.Add(SimpleText(panel, 'Copy To Selected Groups:'),
                  style=RIGHT, dcol=3)


        add_text('E0: ')
        panel.Add(wids['e0'])
        panel.Add((10, 10), dcol=2)
        panel.Add(CopyBtn('e0'), style=RIGHT)


        add_text('R bkg: ')
        panel.Add(wids['rbkg'])
        panel.Add((10, 10), dcol=2)
        panel.Add(CopyBtn('rbkg'), style=RIGHT)

        add_text('k min: ')
        panel.Add(bkg_kmin)
        panel.Add(SimpleText(panel, 'k max:', size=(75, -1)), style=LEFT)
        panel.Add(bkg_kmax)
        panel.Add(CopyBtn('bkg_krange'), style=RIGHT)

        add_text('kweight: ', newrow=True)
        panel.Add(wids['bkg_kweight'])
        panel.Add((10, 10), dcol=2)
        panel.Add(CopyBtn('bkg_kweight'), style=RIGHT)


        add_text('Clamps Low E: ', newrow=True)
        panel.Add( wids['bkg_clamplo'])
        add_text('high E: ',  newrow=False)
        panel.Add( wids['bkg_clamphi'])
        panel.Add(CopyBtn('bkg_clamp'), style=RIGHT)

        panel.Add(HLine(panel, size=(500, 3)), dcol=6, newrow=True)

        panel.Add(SimpleText(panel, ' Fourier transform    ', size=(180, -1),
                             **self.titleopts), dcol=2, style=LEFT, newrow=True)
        panel.Add(SimpleText(panel, 'Copy To Selected Groups:'),
                  style=RIGHT, dcol=3)


        panel.Add(SimpleText(panel, 'k min: '), newrow=True)
        panel.Add(fft_kmin)

        panel.Add(SimpleText(panel, 'k max:', size=(75, -1)), style=LEFT)
        panel.Add(fft_kmax)
        panel.Add(CopyBtn('fft_krange'), style=RIGHT)

        panel.Add(SimpleText(panel, 'k weight : '), newrow=True)
        panel.Add(wids['fft_kweight'])
        panel.Add((10, 10), dcol=2)
        panel.Add(CopyBtn('fft_kweight'), style=RIGHT)


        panel.Add(SimpleText(panel, 'K window : '), newrow=True)
        panel.Add(wids['fft_kwindow'])
        panel.Add(SimpleText(panel, ' dk : '))
        panel.Add(wids['fft_dk'])
        panel.Add(CopyBtn('fft_kwindow'), style=RIGHT)

        panel.Add((10, 10), newrow=True)
        panel.Add(self.wids['is_frozen'], dcol=1, newrow=True)
        panel.Add(saveconf, dcol=4)
        panel.Add((10, 10), newrow=True)
        panel.Add(HLine(self, size=(500, 3)), dcol=8, newrow=True)

        panel.pack()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add((10, 10), 0, LEFT, 3)
        sizer.Add(panel, 1, LEFT, 3)
        pack(self, sizer)
        self.skip_process = False

    def get_config(self, dgroup=None):
        """get and set processing configuration for a group"""
        if dgroup is None:
            dgroup = self.controller.get_group()
        # print("EXAFS panel get_config ", dgroup)
        conf = getattr(dgroup, self.configname, self.get_defaultconfig())
        # print("EXAFS panel get_config ", conf.get('e0', -123.))

        bkg_kmax = conf.get('bkg_kmax', None)
        fft_kmax = conf.get('fft_kmax', None)

        if None in (bkg_kmax, fft_kmax):
            e0 = conf.get('e0', -1)
            emin = min(dgroup.energy)
            if e0 is None or e0 < emin:
                e0 = getattr(dgroup, 'e0',  emin)
            kmax = etok(max(dgroup.energy) - e0)

            if bkg_kmax is None or bkg_kmax < 0:
                conf['bkg_kmax'] = kmax + 0.1
            if fft_kmax is None or fft_kmax < 0:
                conf['fft_kmax'] = kmax - 1

        if dgroup is not None:
            setattr(dgroup, self.configname, conf)
        return conf


    def fill_form(self, dgroup):
        """fill in form from a data group"""
        opts = self.get_config(dgroup)
        self.dgroup = dgroup
        self.skip_process = True
        wids = self.wids
        for attr in ('e0', 'rbkg', 'bkg_kmin', 'bkg_kmax',
                     'bkg_kweight', 'fft_kmin', 'fft_kmax',
                     'fft_kweight', 'fft_dk'):
            val = getattr(dgroup, attr, None)
            if val is None:
                val = opts.get(attr, -1)
            wids[attr].SetValue(val)

        for attr in ('bkg_clamplo', 'bkg_clamphi'):
            wids[attr].SetStringSelection("%d" % opts.get(attr, 0))

        for attr in ('fft_kwindow', 'plotone_op', 'plotsel_op', 'plotalt_op'):
            if attr in opts:
                wids[attr].SetStringSelection(opts[attr])


        frozen = opts.get('is_frozen', False)
        if hasattr(dgroup, 'is_frozen'):
            frozen = dgroup.is_frozen

        self.wids['is_frozen'].SetValue(frozen)
        self._set_frozen(frozen)

        self.skip_process = False

    def read_form(self, dgroup=None):
        "read form, return dict of values"
        skip_save = self.skip_process
        self.skip_process = True

        if dgroup is None:
            dgroup = self.controller.get_group()
        self.dgroup = dgroup
        if dgroup is None:
            return {}
        form_opts = {'group': dgroup.groupname}

        wids = self.wids
        for attr in ('e0', 'rbkg', 'bkg_kmin', 'bkg_kmax',
                     'bkg_kweight', 'fft_kmin', 'fft_kmax',
                     'fft_kweight', 'fft_dk', 'plot_kweight',
                     'plot_kweight_alt', 'plot_voffset'):
            form_opts[attr] = wids[attr].GetValue()

        for attr in ('bkg_clamplo', 'bkg_clamphi'):
            form_opts[attr] = int(wids[attr].GetStringSelection())

        for attr in ('fft_kwindow', 'plotone_op', 'plotsel_op', 'plotalt_op'):
            form_opts[attr] = wids[attr].GetStringSelection()
        time.sleep(0.001)
        conf = self.get_config()

        conf.update(form_opts)

        self.skip_process = skip_save
        return form_opts


    def onSaveConfigBtn(self, evt=None):
        self.read_form()
        self.set_defaultconfig(conf)

    def onCopyParam(self, name=None, evt=None):

        self.read_form()
        conf = self.get_config()
        opts = {}
        def copy_attrs(*args):
            return {a: conf[a] for a in args}

        name = str(name)
        if name in ('e0', 'rbkg', 'bkg_kweight', 'fft_kweight'):
            opts = copy_attrs(name)
        elif name == 'bkg_krange':
            opts = copy_attrs('bkg_kmin', 'bkg_kmax')
        elif name == 'bkg_clamp':
            opts = copy_attrs('bkg_clamplo', 'bkg_clamphi')
        elif name == 'fft_krange':
            opts = copy_attrs('fft_kmin', 'fft_kmax')
        elif name == 'fft_kwindow':
            opts = copy_attrs('fft_kwindow', 'fft_dk')

        for checked in self.controller.filelist.GetCheckedStrings():
            groupname = self.controller.file_groups[str(checked)]
            grp = self.controller.get_group(groupname)
            if grp != self.controller.group and not grp.is_frozen:
                self.update_config(opts, dgroup=grp)

    def _set_frozen(self, frozen):
        try:
            dgroup = self.controller.get_group()
            dgroup.is_frozen = frozen
        except:
            pass

        for attr in ('e0', 'rbkg', 'bkg_kmin', 'bkg_kmax', 'bkg_kweight',
                     'fft_kmin', 'fft_kmax', 'fft_kweight', 'fft_dk',
                     'bkg_clamplo', 'bkg_clamphi', 'fft_kwindow'):
            self.wids[attr].Enable(not frozen)

    def onFreezeGroup(self, evt=None):
        self._set_frozen(evt.IsChecked())


    def onProcess(self, event=None):
        """ handle process events"""
        if self.skip_process:
            return
        self.skip_process = True
        self.read_form()
        self.process(dgroup=self.dgroup)
        self.skip_process = False

        plotter = self.onPlotOne
        if self.last_plot == 'selected':
            plotter = self.onPlotSel
        wx.CallAfter(partial(plotter))

    def process(self, dgroup=None, **kws):
        if dgroup is not None:
            self.dgroup = dgroup

        conf = self.get_config(self.dgroup)
        conf['group'] = self.dgroup.groupname

        conf.update(kws)
        if not 'fft_kwindow' in conf:
            return

        bkgpars = []
        # print(" EXAFS Process #1 e0= ", conf.get('e0', -10.))
        for attr in ('e0', 'rbkg', 'bkg_kmin', 'bkg_kmax',
                     'bkg_kweight', 'bkg_clamplo', 'bkg_clamphi'):
            val = conf.get(attr, 0.0)
            if val is None:
                val = -1.0
            bkgpars.append("%.3f" % val)
        bkgpars = ':'.join(bkgpars)
        # print(" EXAFS Process #2 : ", bkgpars)

        if bkgpars != self.last_process_bkg.get(self.dgroup.groupname, ''):
            self.larch_eval(autobk_cmd.format(**conf))
            self.last_process_bkg[self.dgroup.groupname] = bkgpars
            self.last_process_fft[self.dgroup.groupname] = ''

        fftpars = [conf['fft_kwindow']]
        for attr in ('fft_kmin', 'fft_kmax', 'fft_kweight', 'fft_dk'):
            fftpars.append("%.3f" % conf.get(attr, 0.0))
        fftpars = ':'.join(fftpars)

        if fftpars != self.last_process_fft.get(self.dgroup.groupname, ''):
            self.larch_eval(xftf_cmd.format(**conf))
            self.last_process_fft[self.dgroup.groupname] = fftpars

        setattr(dgroup, self.configname, conf)

    def plot(self, dgroup=None):
        if self.skip_plotting:
            return
        self.onPlotOne(dgroup=dgroup)

    def onPlotOne(self, evt=None, dgroup=None):
        if self.skip_plotting:
            return
        form = self.read_form()
        if len(form) == 0:
            return
        if dgroup is not None:
            self.dgroup = dgroup
            form['group'] = dgroup.groupname
        self.process(dgroup=self.dgroup)
        form['title'] = '"%s"' % self.dgroup.filename

        cmd = PlotCmds[form['plotone_op']] + ", win=1, title={title:s})"
        # 2nd plot
        cmd2 =  PlotCmds[form['plotalt_op']]
        if cmd2 is not None:
            cmd2 = cmd2.replace('plot_kweight', 'plot_kweight_alt')
            cmd2 = cmd2 + ", win=2, title={title:s})"
            cmd = "%s\n%s" % (cmd, cmd2)
            self.controller.get_display(win=2)

        self.larch_eval(cmd.format(**form))
        self.last_plot = 'one'
        self.parent.SetFocus()
        if evt is not None:
            evt.Skip()

    def onPlotSel(self, evt=None):
        if self.skip_plotting:
            return
        group_ids = self.controller.filelist.GetCheckedStrings()
        if len(group_ids) < 1:
            return

        form = self.read_form()
        bcmd = PlotCmds[form['plotsel_op']]
        form['new'] = 'True'
        offset = form['plot_voffset']
        for i, checked in enumerate(group_ids):
            groupname = self.controller.file_groups[str(checked)]
            dgroup = self.controller.get_group(groupname)
            if dgroup is not None:
                form['group'] = dgroup.groupname
                form['label'] = dgroup.filename
                form['offset'] = offset * i
                self.process(dgroup=dgroup)

                extra = """, offset={offset:.3f}, win=1, delay_draw=True,
    label='{label:s}', new={new:s})"""
                cmd = "%s%s" % (bcmd, extra)
                self.larch_eval(cmd.format(**form))
                form['new'] = 'False'

        self.larch_eval("redraw(win=1, show_legend=True)")
        self.last_plot = 'selected'
        self.parent.SetFocus()
        if evt is not None:
            evt.Skip()



#!/usr/bin/env python
"""
Output data file from Epics Step Scan (slightly different from gse_escan)
"""

import time
import numpy as np
from .. import Group

COM1 = '#'
COM2 = '##'
SEP  = '||'   # separater between value, pvname in header
FILETOP = '##Epics StepScan File'

class EpicsScanData(object):
    """
    Holds data as read from a Scan Data File.
    """
    def __init__(self, filename=None, **kws):
        self.filename = filename
        self.extra_pvs = []
        self.comments  = []
        self.column_keys    = []
        self.column_names   = []
        self.column_units   = []
        self.column_pvnames = []
        self.breakpoints    = []
        self.breakpoint_times = []
        self.__arraymap = None
        self.start_time = None
        self.stop_time = None
        self.data = []
        if filename is not None:
            self.read(filename)

    def get_data(self, key, fold_breakpoints=False):
        """get positioner or detector array either by key, name, or index
        key can be any of (case-insensitive):
            column number, starting at 0
            column label, 'p1', 'd1', etc
            name of column
            pvname of column
        """
        # cache lower-case version of all keys and names
        if self.__arraymap is None:
            self.__arraymap = {}
            for a in (self.column_keys, self.column_names, self.column_pvnames):
                for i, nam in enumerate(a):
                    self.__arraymap[nam.lower()] = i
        #
        if isinstance(key, int) and key > -1 and key < len(self.column_keys):
            icol = key
        else:
            icol = self.__arraymap.get(key.lower(), None)

        if icol is None:
            print( 'cannot find column %s' % repr(key))
            return None
        return self.data[icol]

    def read(self, filename=None):
        if filename is not None:
            self.filename = filename
        try:
            fh = open(self.filename, 'r')
        except IOError:
            print( 'cannot open file %s for read' % self.filename)
            return
        lines = fh.readlines()
        line0 = lines.pop(0)
        if not line0.startswith(FILETOP):
            print( '%s is not a valid Epics Scan file' % self.filename)
            return

        def split_header(line):
            w = [k.strip() for k in line.replace('#', '').split(':', 1)]
            if len(w) == 1:
                w.append('')
            return w
        mode = None
        extras = {}
        modes = {'Time': 'comment', '----': 'data',
                 'Legend Start': 'legend', 'Legend End': 'legend_end',
                 'ExtraPVs Start': 'extras', 'ExtraPVs End': 'extras_end'}
        for line in lines:
            if line.startswith(COM1):
                key, val = split_header(line[:-1])
                if key.startswith('----'): key = '----'
                if key in modes:
                    mode = modes[key]
                    if mode == 'comment':
                        self.stop_time  = val
                        if self.start_time is None:
                            self.start_time = val
                    elif mode == 'extras':
                        self.breakpoints.append(len(self.data))
                        self.breakpoint_times.append(self.stop_time)
                    elif mode == 'extras_end':
                        self.extra_pvs.append(extras)
                        extras = {}
                    continue
                if mode == 'comment':
                    cmt = line[:-1].strip()
                    if cmt.startswith('#'):  cmt = line[1:].strip()
                    self.comments.append(cmt)
                elif mode in ('legend', 'extras'):
                    words = [w.strip() for w in val.split(SEP)]
                    if len(words) == 1: words.append('')
                    if mode == 'extras':
                        extras[key] = (words[0], words[1])
                    else:
                        if len(words) == 2: words.append('')
                        self.column_keys.append(key)
                        self.column_names.append(words[0])
                        self.column_units.append(words[1])
                        self.column_pvnames.append(words[2])

            else: # data!
                self.data.append([float(i) for i in line[:-1].split()])
        #
        self.comments = '\n'.join(self.comments)
        self.data = np.array(self.data).transpose()

def read_stepscan(fname, _larch=None, **kws):
    """read Epics Step Scan file to larch Group"""

    scan = EpicsScanData(fname)
    group = Group()
    group.__name__ ='Epics Step Sscan Data file %s' % fname
    for key, val in scan.__dict__.items():
        if not key.startswith('_'):
            setattr(group, key, val)

    group.get_data = scan.get_data
    return group



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
RIXS data object
================
"""
import numpy as np
from scipy.interpolate import griddata
from silx.io.dictdump import (dicttoh5, h5todict)

from larch.math.gridxyz import gridxyz
from larch.utils.logging import getLogger
_logger = getLogger('rixsdata')  #: module logger


def _tostr(arr):
    """Numpy array to string"""
    return arr.tostring().decode()


class RixsData(object):
    """RIXS plane object"""

    #: loaded from dictionary/HDF5 -> self.load_from_h5()
    sample_name = 'Unknown'
    counter_all, counter_signal, counter_norm = None, None, None
    _x, _y, _z = None, None, None
    ene_in, ene_out, rixs_map = None, None, None
    ene_et, rixs_et_map = None, None
    ene_grid, ene_unit = None, None

    #: setted by self.crop()
    ene_in_crop, ene_out_crop, rixs_map_crop = None, None, None

    grid_method = 'nearest'
    grid_lib = 'scipy'

    _plotter = None

    def __init__(self, name=None, logger=None):
        """Constructor"""

        self.__name__ = name or 'RixsData_{0}'.format(hex(id(self)))
        self._logger = logger or _logger

    def _init_axis_labels(self, unit=None):
        try:
            unit.decode()
        except AttributeError:
            pass
        self.ene_in_label = 'Incoming energy ({0})'.format(unit)
        self.ene_out_label = 'Emitted energy ({0})'.format(unit)
        self.ene_et_label = 'Energy transfer ({0})'.format(unit)

    def load_from_dict(self, rxdict):
        """Load RIXS data from a dictionary

        Parameters
        ----------
        rxdict : dict
            Minimal required structure
            {
             'writer_version': '1.5',
             'sample_name': str,
             '_x': 1D array,
             '_y': 1D array,
             '_z': 1D array,
            }

        Return
        ------
        None, set attributes: self.*
        """
        self.__dict__.update(rxdict)
        self._init_axis_labels(unit=_tostr(self.ene_unit))
        self.grid_rixs_from_col()

    def load_from_h5(self, fname):
        """Load RIXS from HDF5 file"""
        rxdict = h5todict(fname)
        if not ('writer_version' in rxdict.keys()):
            self._logger.error("Key 'writer_version' not found")
            return
        if not (_tostr(rxdict['writer_version']) == '1.5'):
            self._logger.warning('Data format not understood')
            return
        rxdict['sample_name'] = _tostr(rxdict['sample_name'])
        self.load_from_dict(rxdict)
        self._logger.info("RIXS map loaded from file: {0}".format(fname))

    def save_to_h5(self, fname):
        """Dump dictionary representation to HDF5 file"""
        dicttoh5(self.__dict__)
        self._logger.info("RixsData saved to {0}".format(fname))

    def crop(self, crop_area, yet=False):
        """Crop the plane in a given range

        Parameters
        ----------
        crop_area : tuple
            (x1, y1, x2, y2) : floats
            x1 < x2 (ene_in)
            y1 < y2 (if yet=False: ene_out, else: ene_et)
        yet: bool
            if True: y1, y2 are given in energy transfer [False]
        """
        self._crop_area = crop_area
        x1, y1, x2, y2 = crop_area
        _xystep = self.ene_grid or 0.1
        _method = self.grid_method or 'nearest'

        _nxpts = int((x2-x1)/_xystep)
        _xcrop = np.linspace(x1, x2, num=_nxpts)

        if yet:
            _netpts = int((y2-y1)/_xystep)
            _ymin = x2-y2
            _ymax = x1-y1
            _nypts = int((_ymax-_ymin)/_xystep)
            _etcrop = np.linspace(y1, y2, num=_netpts)
            _ycrop = np.linspace(_ymin, _ymax, num=_nypts)
        else:
            _nypts = int((y2-y1)/_xystep)
            _etmin = x1-y2
            _etmax = x2-y1
            _netpts = int((_etmax-_etmin)/_xystep)
            _etcrop = np.linspace(_etmin, _etmax, num=_netpts)
            _ycrop = np.linspace(y1, y2, num=_nypts)

        _xx, _yy = np.meshgrid(_xcrop, _ycrop)
        _exx, _et = np.meshgrid(_xcrop, _etcrop)
        _logger.info("Gridding data...")
        _zzcrop = griddata((self._x, self._y), self._z, (_xx, _yy), method=_method)
        _ezzcrop = griddata((self._x, self._x-self._y), self._z, (_exx, _et), method=_method)

        self.ene_in_crop = _xcrop
        self.ene_out_crop = _ycrop
        self.ene_et_crop = _etcrop
        self.rixs_map_crop = _zzcrop
        self.rixs_et_map_crop = _ezzcrop

    def grid_rixs_from_col(self):
        """Grid RIXS map from XYZ"""
        _lib = self.grid_lib or 'scipy'
        _method = self.grid_method or 'nearest'
        _xystep = self.ene_grid or 0.1
        self.ene_in, self.ene_out, self.rixs_map = gridxyz(self._x,
                                                           self._y,
                                                           self._z,
                                                           xystep=_xystep,
                                                           lib=_lib,
                                                           method=_method)
        self._et = self._x - self._y
        _, self.ene_et, self.rixs_et_map = gridxyz(self._x, self._et, self._z,
                                                   xystep=_xystep,
                                                   lib=_lib,
                                                   method=_method)

    def norm(self):
        """Simple map normalization to max-min"""
        self.rixs_map_norm = self.rixs_map/(np.nanmax(self.rixs_map)-np.nanmin(self.rixs_map))


if __name__ == '__main__':
    pass



#!/usr/bin/env python
"""
  XAFS MBACK normalization algorithms.
"""
import numpy as np
from scipy.special import erfc

from xraydb import (xray_edge, xray_line, xray_lines,
                    f1_chantler, f2_chantler, guess_edge,
                    atomic_number, atomic_symbol)
from lmfit import Parameter, Parameters, minimize

from larch import Group, isgroup, parse_group_args

from larch.math import index_of, index_nearest, remove_dups, remove_nans2

from .xafsutils import set_xafsGroup
from .pre_edge import find_e0, preedge


MAXORDER = 6

def find_xray_line(z, edge):
    """
    Finds most intense X-ray emission line energy for a given element and edge.
    """
    intensity = 0
    line      = ''
    for key, value in xray_lines(z).items() :
        if value.initial_level == edge.upper():
            if value.intensity > intensity:
                intensity = value.intensity
                line      = key
    return xray_line(z, line[:-1])

def match_f2(p, en=0, mu=1, f2=1, e0=0, em=0, weight=1, theta=1, order=None,
             leexiang=False):
    """
    Objective function for matching mu(E) data to tabulated f''(E) using the MBACK
    algorithm and, optionally, the Lee & Xiang extension.
    """
    pars = p.valuesdict()
    eoff  = en - e0

    norm = p['a']*erfc((en-em)/p['xi']) + p['c0'] # erfc function + constant term
    for i in range(order):                        # successive orders of polynomial
        attr = 'c%d' % (i + 1)
        if attr in p:
            norm += p[attr] * eoff**(i + 1)
    func = (f2 + norm - p['s']*mu) * theta / weight
    if leexiang:
        func = func / p['s']*mu
    return func


def mback(energy, mu=None, group=None, z=None, edge='K', e0=None, pre1=None, pre2=-50,
          norm1=100, norm2=None, order=3, leexiang=False, tables='chantler', fit_erfc=False,
          return_f1=False, _larch=None):
    """
    Match mu(E) data for tabulated f''(E) using the MBACK algorithm and,
    optionally, the Lee & Xiang extension

    Arguments
    ----------
      energy:     array of x-ray energies, in eV.
      mu:         array of mu(E).
      group:      output group.
	  z:          atomic number of the absorber.
	  edge:       x-ray absorption edge (default 'K')
      e0:         edge energy, in eV.  If None, it will be determined here.
      pre1:       low E range (relative to e0) for pre-edge region.
      pre2:       high E range (relative to e0) for pre-edge region.
      norm1:      low E range (relative to e0) for post-edge region.
      norm2:      high E range (relative to e0) for post-edge region.
      order:      order of the legendre polynomial for normalization.
	              (default=3, min=0, max=5).
      leexiang:   boolean (default False)  to use the Lee & Xiang extension.
      tables:     tabulated scattering factors: 'chantler' [deprecated]
      fit_erfc:   boolean (default False) to fit parameters of error function.
      return_f1:  boolean (default False) to include the f1 array in the group.


    Returns
    -------
      None

    The following attributes will be written to the output group:
      group.f2:            tabulated f2(E).
      group.f1:            tabulated f1(E) (if 'return_f1' is True).
      group.fpp:           mback atched spectrum.
	  group.edge_step:     edge step of spectrum.
	  group.norm:          normalized spectrum.
      group.mback_params:  group of parameters for the minimization.

    Notes:
        Chantler tables is now used, with Cromer-Liberman no longer supported.
    References:
      * MBACK (Weng, Waldo, Penner-Hahn): http://dx.doi.org/10.1086/303711
      * Lee and Xiang: http://dx.doi.org/10.1088/0004-637X/702/2/970
      * Cromer-Liberman: http://dx.doi.org/10.1063/1.1674266
      * Chantler: http://dx.doi.org/10.1063/1.555974
    """
    order = max(min(order, MAXORDER), 0)

    ### implement the First Argument Group convention
    energy, mu, group = parse_group_args(energy, members=('energy', 'mu'),
                                         defaults=(mu,), group=group,
                                         fcn_name='mback')
    if len(energy.shape) > 1:
        energy = energy.squeeze()
    if len(mu.shape) > 1:
        mu = mu.squeeze()

    if _larch is not None:
        group = set_xafsGroup(group, _larch=_larch)

    energy = remove_dups(energy)
    if e0 is None or e0 < energy[1] or e0 > energy[-2]:
        e0 = find_e0(energy, mu, group=group)

    print(e0)
    ie0 = index_nearest(energy, e0)
    e0 = energy[ie0]

    pre1_input = pre1
    norm2_input = norm2

    if pre1 is None:  pre1  = min(energy) - e0
    if norm2 is None: norm2 = max(energy) - e0
    if norm2 < 0:     norm2 = max(energy) - e0 - norm2
    pre1  = max(pre1,  (min(energy) - e0))
    norm2 = min(norm2, (max(energy) - e0))

    if pre1 > pre2:
        pre1, pre2 = pre2, pre1
    if norm1 > norm2:
        norm1, norm2 = norm2, norm1

    p1 = index_of(energy, pre1+e0)
    p2 = index_nearest(energy, pre2+e0)
    n1 = index_nearest(energy, norm1+e0)
    n2 = index_of(energy, norm2+e0)
    if p2 - p1 < 2:
        p2 = min(len(energy), p1 + 2)
    if n2 - n1 < 2:
        p2 = min(len(energy), p1 + 2)

    ## theta is a boolean array indicating the
	## energy values considered for the fit.
    ## theta=1 for included values, theta=0 for excluded values.
    theta            = np.zeros_like(energy, dtype='int')
    theta[p1:(p2+1)] = 1
    theta[n1:(n2+1)] = 1

    ## weights for the pre- and post-edge regions, as defined in the MBACK paper (?)
    weight            = np.ones_like(energy, dtype=float)
    weight[p1:(p2+1)] = np.sqrt(np.sum(weight[p1:(p2+1)]))
    weight[n1:(n2+1)] = np.sqrt(np.sum(weight[n1:(n2+1)]))

    ## get the f'' function from CL or Chantler
    f1 = f1_chantler(z, energy)
    f2 = f2_chantler(z, energy)
    group.f2 = f2
    if return_f1:
        group.f1 = f1

    em = find_xray_line(z, edge).energy # erfc centroid

    params = Parameters()
    params.add(name='s',  value=1.0,  vary=True)  # scale of data
    params.add(name='xi', value=50.0, vary=False, min=0) # width of erfc
    params.add(name='a',  value=0.0, vary=False)  # amplitude of erfc
    if fit_erfc:
        params['a'].vary  = True
        params['a'].value = 0.5
        params['xi'].vary  = True

    for i in range(order+1): # polynomial coefficients
        params.add(name='c%d' % i, value=0, vary=True)

    out = minimize(match_f2, params, method='leastsq',
                   gtol=1.e-5, ftol=1.e-5, xtol=1.e-5, epsfcn=1.e-5,
                   kws = dict(en=energy, mu=mu, f2=f2, e0=e0, em=em,
                              order=order, weight=weight, theta=theta, leexiang=leexiang))

    opars = out.params.valuesdict()
    eoff = energy - e0

    norm_function = opars['a']*erfc((energy-em)/opars['xi']) + opars['c0']
    for i in range(order):
        attr = 'c%d' % (i + 1)
        if attr in opars:
            norm_function  += opars[attr]* eoff**(i + 1)

    group.e0 = e0
    group.fpp = opars['s']*mu - norm_function
    # calculate edge step and normalization from f2 + norm_function
    pre_f2 = preedge(energy, group.f2+norm_function, e0=e0, pre1=pre1,
	         pre2=pre2, norm1=norm1, norm2=norm2, nnorm=2, nvict=0)
    group.edge_step = pre_f2['edge_step'] / opars['s']
    group.norm = (opars['s']*mu -  pre_f2['pre_edge']) / pre_f2['edge_step']
    group.mback_details = Group(params=opars, pre_f2=pre_f2,
                                f2_scaled=opars['s']*f2,
                                norm_function=norm_function)



def f2norm(params, en=1, mu=1, f2=1, weights=1):

    """
    Objective function for matching mu(E) data to tabulated f''(E)
    """
    p = params.valuesdict()
    model = (p['offset'] + p['slope']*en + f2) * p['scale']
    return weights * (model - mu)


def mback_norm(energy, mu=None, group=None, z=None, edge='K', e0=None,
               pre1=None, pre2=None, norm1=None, norm2=None, nnorm=None, nvict=1,
               _larch=None):
    """
    simplified version of MBACK to Match mu(E) data for tabulated f''(E)
    for normalization

    Arguments:
      energy, mu:  arrays of energy and mu(E)
      group:       output group (and input group for e0)
      z:           Z number of absorber
      e0:          edge energy
      pre1:        low E range (relative to E0) for pre-edge fit
      pre2:        high E range (relative to E0) for pre-edge fit
      norm1:       low E range (relative to E0) for post-edge fit
      norm2:       high E range (relative to E0) for post-edge fit
      nnorm:       degree of polynomial (ie, nnorm+1 coefficients will be
                   found) for post-edge normalization curve fit to the
                   scaled f2. Default=1 (linear)

    Returns:
      group.norm_poly:     normalized mu(E) from pre_edge()
      group.norm:          normalized mu(E) from this method
      group.mback_mu:      tabulated f2 scaled and pre_edge added to match mu(E)
      group.mback_params:  Group of parameters for the minimization

    References:
      * MBACK (Weng, Waldo, Penner-Hahn): http://dx.doi.org/10.1086/303711
      * Chantler: http://dx.doi.org/10.1063/1.555974
    """
    ### implement the First Argument Group convention
    energy, mu, group = parse_group_args(energy, members=('energy', 'mu'),
                                         defaults=(mu,), group=group,
                                         fcn_name='mback')
    if len(energy.shape) > 1:
        energy = energy.squeeze()
    if len(mu.shape) > 1:
        mu = mu.squeeze()

    if _larch is not None:
        group = set_xafsGroup(group, _larch=_larch)
    group.norm_poly = group.norm*1.0

    if z is not None:              # need to run find_e0:
        e0_nominal = xray_edge(z, edge).energy
    if e0 is None:
        e0 = getattr(group, 'e0', None)
        if e0 is None:
            find_e0(energy, mu, group=group)
            e0 = group.e0

    atsym = None
    if z is None or z < 2:
        atsym, edge = guess_edge(group.e0)
        z = atomic_number(atsym)
    if atsym is None and z is not None:
        atsym = atomic_symbol(z)

    if getattr(group, 'pre_edge_details', None) is None:  # pre_edge never run
        preedge(energy, mu, pre1=pre1, pre2=pre2, nvict=nvict,
                norm1=norm1, norm2=norm2, e0=e0, nnorm=nnorm)

    mu_pre = mu - group.pre_edge
    f2 = f2_chantler(z, energy)

    weights = np.ones(len(energy))*1.0

    if norm2 is None:
        norm2 = max(energy) - e0

    if norm2 < 0:
        norm2 = max(energy) - e0  - norm2

    # avoid l2 and higher edges
    if edge.lower().startswith('l'):
        if edge.lower() == 'l3':
            e_l2 = xray_edge(z, 'L2').energy
            norm2 = min(norm2,  e_l2-e0)
        elif edge.lower() == 'l2':
            e_l2 = xray_edge(z, 'L1').energy
            norm2 = min(norm2,  e_l1-e0)

    ipre2 = index_of(energy, e0+pre2)
    inor1 = index_of(energy, e0+norm1)
    inor2 = index_of(energy, e0+norm2) + 1


    weights[ipre2:] = 0.0
    weights[inor1:inor2] = np.linspace(0.1, 1.0, inor2-inor1)

    params = Parameters()
    params.add(name='slope',   value=0.0,    vary=True)
    params.add(name='offset',  value=-f2[0], vary=True)
    params.add(name='scale',   value=f2[-1], vary=True)

    out = minimize(f2norm, params, method='leastsq',
                   gtol=1.e-5, ftol=1.e-5, xtol=1.e-5, epsfcn=1.e-5,
                   kws = dict(en=energy, mu=mu_pre, f2=f2, weights=weights))

    p = out.params.valuesdict()

    model = (p['offset'] + p['slope']*energy + f2) * p['scale']

    group.mback_mu = model + group.pre_edge

    pre_f2 = preedge(energy, model, nnorm=nnorm, nvict=nvict, e0=e0,
                     pre1=pre1, pre2=pre2, norm1=norm1, norm2=norm2)

    step_new = pre_f2['edge_step']

    group.edge_step_poly  = group.edge_step
    group.edge_step_mback = step_new
    group.norm_mback = mu_pre / step_new


    group.mback_params = Group(e0=e0, pre1=pre1, pre2=pre2, norm1=norm1,
                               norm2=norm2, nnorm=nnorm, fit_params=p,
                               fit_weights=weights, model=model, f2=f2,
                               pre_f2=pre_f2, atsym=atsym, edge=edge)

    if (abs(step_new - group.edge_step)/(1.e-13+group.edge_step)) > 0.75:
        print("Warning: mback edge step failed....")
    else:
        group.edge_step = step_new
        group.norm       = group.norm_mback



#!/usr/bin/env python
"""
Use Epics Channel Access
"""

HAS_PYEPICS = False

try:
    import epics
    HAS_PYEPICS = True
except ImportError:
    HAS_PYEPICS = False

from . import larchscan

def pv_fullname(name):
    """ make sure an Epics PV name ends with .VAL or .SOMETHING!

    Parameters
    ----------
       pvname:   name of PV

    Returns
    -------
       string with full PV name

    """
    name = str(name)
    if '.' not in name:
        name = "%s.VAL" % name
    return name

def pv_units(pv, default=''):
    """get units for pv object, with optional default value

    Parameters
    ----------
       pv:       pv object (created by PV())
       default:  string value for default units

    Returns
    -------
       string with units

    """
    try:
        units = pv.units
    except:
        units = ''
    if units in (None, ''):
        units = default
    return units



def nullfcn(*args,  **kws):
    "pyepics is not installed"
    return None

caget = caput = cainfo = PV = nullfcn
__DOC__ = "pyepics is not installed"
epics_exports = {}
scan_exports = {}

_larch_name = '_epics'
_larch_builtins = {'_epics': {}}


if HAS_PYEPICS:
    from .xrf_detectors import Epics_MultiXMAP, Epics_Xspress3
    from .xrfcontrol import EpicsXRFApp

    caget = epics.caget
    caput = epics.caput
    cainfo = epics.cainfo
    PV = epics.get_pv
    __DOC__ = """
Functions for Epics Channel Access, using pyepics interface.
For further details, consult the pyepics documentation

  Functions
  -------------
  caget(pvname, as_string=False, count=None, timeout=None)
           get and return value for an Epics Process Variable
           with options to ensure return value is a string,
           to limit the count for array values, or to specify
           a maximum time to wait for a value.
  caput(pvname, value, wait=False, timeout=60)
           put a value to an Epics Process Variable, optionally
           waiting to return until record is fully processed,
           with a timeout value.
  cainfo(pvname, print_out=True)
           print out information about Process Variable. with
           print_out = False, returns string of information.
  PV(pvname)
           create a Process Variable object with get()/put()
           and several other methods.
  pv_units(pvname)
           return units for PV
  pv_fullname(pvname)
           return fullname (that is, including '.VAL' if needed) for PV
"""

    _larch_builtins = {'_epics': dict(PV=PV, caget=caget, caput=caput,
                                      cainifo=cainfo, pv_units=pv_units,
                                      pv_fullname=pv_fullname)}

    if larchscan.HAS_EPICSSCAN:
        _larch_builtins['_scan'] = dict(scan_from_db=larchscan.scan_from_db,
                                        connect_scandb=larchscan.connect_scandb,
                                        do_scan=larchscan.do_scan,
                                        do_slewscan=larchscan.do_scan,
                                        get_dbinfo=larchscan.get_dbinfo,
                                        set_dbinfo=larchscan.set_dbinfo)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Handling loggers
"""
import sys
import logging

# import tempfile

# set up default logging configureation
_default_format = '[%(name)-s] %(levelname)-s : %(message)s'
_default_format_file = '%(asctime)s [%(name)-s] %(levelname)-s : %(message)s'
_default_datefmt = '%Y-%m-%d %H:%M'
_levels = {'DEBUG': logging.DEBUG,
           'INFO': logging.INFO,
           'WARNING': logging.WARNING,
           'ERROR': logging.ERROR,
           'FATAL': logging.FATAL,
           'CRITICAL': logging.CRITICAL}

# # needs colorlog, not enabled yet!
# _default_colors = {'DEBUG': 'cyan',
#                    'INFO': 'green',
#                    'WARNING': 'yellow',
#                    'ERROR':    'red',
#                    'CRITICAL': 'red'}
#

_log = False


def logging_basicConfig(level='INFO'):
    """logging basic configuration"""
    global _log
    logging.basicConfig(level=_levels[level],
                        format=_default_format,
                        datefmt=_default_datefmt)
                        # filename='{0}.log'.format(tempfile.mktemp()),
                        # filemode='w')
    _log = True


def getConsoleHandler():
    """Default console handler"""
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(logging.Formatter(fmt=_default_format))
    return console_handler


def getFileHandler(filename, mode='a'):
    """Default console handler"""
    file_handler = logging.FileHandler(filename, mode=mode)
    file_handler.setFormatter(logging.Formatter(fmt=_default_format_file,
                                                datefmt=_default_datefmt))
    return file_handler


def getLogger(name, level='INFO'):
    """Utility function to get the logger with customization

    .. warning:: NOT WORKING AS EXPECTED -> FIXME!!!

    Parameters
    ----------
    name : str
        name of the logger
    level : str (optional)
        logging level ['INFO']
    """
    if not _log:
        logging_basicConfig(level=level)
    logger = logging.getLogger(name)
    logger.setLevel(_levels[level])

    if (logger.hasHandlers()):
        logger.handlers.clear()
        logger.debug("custom logger setup > clear handlers")
    logger.addHandler(getConsoleHandler())
    """Clear existing handlers and add current one"""

    logger.propagate = False
    logger.debug("custom logger setup > finished")

    return logger


def test_logger(level='DEBUG'):
    """Test custom logger"""
    logger = getLogger('test', level=level)
    logger.debug('a debug message')
    logger.info('an info message')
    logger.warning('a warning message')
    logger.error('an error message')
    logger.critical('a critical message')
    import tempfile
    flog = tempfile.mktemp(prefix='test_logger_', suffix='.log')
    logger.addHandler(getFileHandler(flog))
    logger.info(f'added file handler -> {flog}')
    logger.info('testing again all log levels:')
    logger.debug('a debug message')
    logger.info('an info message')
    logger.warning('a warning message')
    logger.error('an error message')
    logger.critical('a critical message')



if __name__ == '__main__':
    test_logger()



"""
Deadtime calculations

Authors/Modifications:
----------------------
* T. Trainor (tptrainor@alaska.edu), 6-10-2006

Description:
------------
The objective is to calculate a factor 'cor' that will provide
a deadtime correction factor, and give corrected counts via:

    counts_corrected = counts * cor

Background:
------------
A correction factor can be defined as:

    cor = (icr_t/ocr_s)*(rt/lt)

Here icr_t = true input count rate (actual counts hitting the detector
per detector live time - ie icr is a rate)

ocr_s is the output rate from detector slow channel:  ocr_s = TOC_s/lt
TOC_s is the total counts output from the slow filter (~ the total counts output
by the detector).

rt and lt are the detector real time and live time respectively. real time is
the requested counting time (=time elapsed from detector start to detector stop).
live time is the amount of time the detector is active.  e.g. lt/rt*100% is the
percent of the counting time that the detector is actually live.

icr_t is an uknown quantity.  It can be determined or approximated in a few
different ways.

A) icr_t may be determined by inverting the following:

    ocr_f = icr_t * exp( -icr_t * t_f)

Here ocr_f is the fast count rate.  ocr_f = TOC_f/lt, TOC_f are the total counts
output from the fast filter.

t_f is the fast filter deadtime.  If the detector reports TOC_f and t_f is known
(from fit to a deatime curve) then icr_t may be calculated from the above.

Note: a detector/file may report InputCounts or ICR.
This is likely = ocr_f rather than icr_t.

Note: The fast filter deatime is much less substantial than the slow filter
deadtime.  Therefore for low total count rates the approximation icr_t ~ ocr_f
may be reasonable.

B) alternatively icr_t may be determined by inverting the following:

    ocr_s = icr_t * exp( -icr_t * t_s)

Here ocr_s is the slow count rate.  ocr_s =TOC_s/lt, TOC_s are the total counts
output from the slow filter.

t_s is the slow filter deadtime.  If the detector reports TOC_s and t_s is known
(from fit to a deatime curve) then icr_t may be calculated from the above.
If the detector does not report TOC_s ocr_s may be approximated from:

    ocr_s = ( Sum_i ( cnts_i) ) / lt ~ TOC_s/lt

Note the above discussion referes primarily to digital couting electronics,
using slow and fast filters. Analog electronics using an SCA may be corrected
using a similiar procedure.  Here the analog amplifier may
report total input counts.  In this case icr_t ~ icr_a (icr amplifier).


Measuring t's:
--------------

To perform a deadtime correction the characteristic deadtimes must be determined.
In each case (t_f or t_s) a deadtime scan is run, where the input counts are
varied (eg scanning a slit) and output counts monitored.

If a measure of icr_t is reported (or approximated ie = ocr_f or icr_a), then
t_f may be deterimined directly from a fit of
  * x = icr_t
  * y = ocr_s

More likely a detector not suffering from deadtime will be used as a proxy for
icr_t (eg mon -> ion chamber)

In this case plot:
  * x = mon/rt
  * y = ocr_f   -> TOC_f/lt reported in file (may be called icr or total counts)
  * icr_t = a*mon/rt + off
  * fit varying t_f and a and off (slope and offset btwn mon and icr_t)
      ocr_f = icr_t*exp(-icr_t*t_f)

If the detector does not report ocr_f, then use slow counts to correct
(or just sum all the actual output counts):
  * x = mon/rt
  * y = ocr_s   -> TOC_s/lt ~ ( Sum_i ( cnts_i) ) / lt
  * icr_t = a*mon/rt + off
  * fit varying t_s and a and off (slope and offset btwn mon and icr_t)
         ocr_s = icr_t*exp(-icr_t*t_s)

Summary of correction methods:
------------------------------
To apply a deadtime correction then depends on what data can be accessed from
the detector.

1) detector reports actual icr_t (or reports ocr_f and total counts are low so
   ocr_f ~icr_t). Then the correction may be calculated directly from
   detector numbers

2) The detector reports ocr_f and t_f has been determined.  icr_t is calculated
   from the saturation equation to use in the correction.

3) The detector does not report ocr_f, rather ocr_s is reported (or approximated
   from total counts) and t_f has been determined.  icr_t is again calculated
   from the saturation equation and used for the correction.
   This is probably the most straightforward method since icr_s can be
   approximated directly from the sum of the counts in the detector (norm by lt),
   and this approach should work for analog and digital electronics.

4) icr_t is uknown or icr_t ~ ocr_s then just assume that icr_t/ocr_s = 1
   (ie just correct for livetime effects)

Example:
--------
# given the arrays mon (ion chamber), rt and lt (each len = npts) and
# the (numpy) multidimensional array for counts (e.g. nptsx2048)
>>ocr = counts.sum(1)/lt
>>x = mon/lt
>>(params,msg) = fit(x,ocr)
>>tau = params[0]
>>a   = params[1]
>>print('a_fit= ',a,' tau_fit=', tau)
# corrected counts
>>counts_cor = num.ones(counts.shape)
>>ocr_cor = num.ones(mon.shape)
>>for j in range(len(counts)):
...>icr = calc_icr(ocr[j],tau)
...>cor = correction_factor(rt[j],lt[j],icr[j],ocr[j])
...>counts_cor[j] = counts[j]*cor
...>ocr_cor[j] = counts_cor[j].sum()/lt
>>pyplot.plot(x,ocr)
>>pyplot.plot(x,ocr_cor)
"""

##############################################################################

import numpy as np
import scipy
from scipy.optimize import leastsq
from scipy.stats import linregress

E_INV = np.exp(-1)

##############################################################################
def correction_factor(rt, lt, icr=None, ocr=None):
    """
    Calculate the deadtime correction factor.

    Parameters:
    -----------
    * rt  = real time, time the detector was requested to count for
    * lt  = live time, actual time the detector was active and
            processing counts
    * icr = true input count rate (TOC_t/lt, where TOC_t = true total counts
            impinging the detector)
    * ocr = output count rate (TOC_s/lt, where TOC_s = total processed
            {slow filter for dxp} counts output by the detector)

    If icr and/or ocr are None then only lt correction is applied

    Outputs:
    -------
    * cor = (icr/ocr)*(rt/lt)
      the correction factor. the correction is applied as:
        corrected_counts = counts * cor
    """
    if icr is not None and ocr is not None:
        return (icr/ocr)*(rt/lt)
    return (rt/lt)

def correct_data(data,rt,lt,icr=None,ocr=None):
    """
    Apply deatime correction to data
    """
    cor = correction_factor(rt, lt, icr, ocr)
    return data * cor

def calc_icr(ocr, tau):
    """
    Calculate the true icr from a given ocr and corresponding deadtime factor
    tau using a Newton-Raphson algorithm to solve the following expression.

        ocr = icr * exp(-icr*tau)

    Returns None if the loop cannot converge

    Note below could be improved!
    """
    # error checks
    if ocr is None or tau is None or ocr <= 0:
        return None

    # here assume if tau = 0, icr=ocr ie icr/ocr =1
    if tau <=0:
        return ocr

    # max_icr is icr val at top of deadtime curve
    # max_ocr is the corresponding ocr value
    # we cannot correct the data if ocr > ocr_max
    max_icr = 1/tau
    max_ocr = max_icr*E_INV
    if ocr > max_ocr:
        print( 'ocr exceeds maximum correctible value of %g cps' % max_ocr)
        return None

    # Newton loop
    icr0 = ocr
    cnt = 0
    while cnt < 100:
        cnt += 1
        delta = (ocr*np.exp(icr0*tau) - icr0) / (icr0*tau - 1)
        if  abs(delta) < 0.01:
            icr = icr0
            break
        else:
            icr0 = icr0 - delta
            if icr0 > max_icr:
                # went over the top, we assume that
                # the icr is less than 1/tau
                icr0 = 1.1 * ocr

    if cnt >= 100:
        print( 'Warning: icr calculation failed to converge')
        icr = None
    return icr

##############################################################################
def fit_deadtime(mon, ocr, offset=True):
    """
    Fit a deatime curve and return optimized value of tau

    This fits the data to the following:
          x = icr_t = a*mon + off -> linear relation btwn icr and mon
          y = ocr   -> TOC/lt
     fit varying 'tau', and 'a' and 'off' (slope and offset btwn mon and icr_t)
         ocr = icr_t*exp(-icr_t*tau)

    This function is appropriate for fitting either slow or fast ocr's.
    If mon is icr_t the the optimized 'a' should be ~1.

    Parameters:
    -----------
    * mon is an array from a linear detector.  This should be in counts/sec
      (ie mon_cnts/scaler_count_time)

    * ocr is an array corresponding to the output count rate (either slow or fast).
      This should be in counts per second where
         ocr = TOC/lt,
      TOC is the total output counts and lt is the detector live time

    * If offset = False, off = 0.0 and only tau and a are returned

    Example:
    --------
    >>params = fit_deadtime(mon/time, ocr)
    >>tau = params[0]
    >>a   = params[1]
    >>off = params[2]

    """
    mon  = np.array(mon,dtype=float)
    ocr = np.array(ocr,dtype=float)

    npts = len(mon)
    if len(ocr) != npts:
        return None

    # make a guess at tau, assume max(ocr) is the top of the deadtime curve
    tau = E_INV/max(ocr)

    # make a guess at linear params, ie if the detector is linear
    #    ocr = icr = a*mon + off
    # assume that mon and ocr are arranged in ascending order, and the first 10%
    # have little deadtime effects (or maybe a bit so scale up the average 10-20%).

    idx = max(int(0.10*npts), 3)
    try:
        xx = linregress(mon[0:idx],ocr[0:idx])
        a   = xx[0]
        off = xx[1]
    except:
        mon_avg = mon[:idx].sum()/idx
        ocr_avg = ocr[:idx].sum()/idx
        a   = 1.2*ocr_avg/mon_avg
        off = 0.0
    if offset:
        params = (tau, a, off)
    else:
        params = (tau,a)
    return leastsq(deadtime_residual, params, args = (mon, ocr, offset))

def calc_ocr(params, mon, offset):
    """ compute ocr from params"""
    tau, a = params[0], params[1]
    off = 0
    if offset:
        off = params[2]
    return (a*mon+off) * np.exp(-tau*(a*mon+off))

def deadtime_residual(params, mon, ocr, offset):
    """ compute residual """
    return ocr - calc_ocr(params, mon, offset)

##############################################################################
if __name__ == '__main__':
    # test fit
    mon  = 10000. * np.arange(500.0)
    a   = 0.1
    tau = 0.00001
    print( 'a= ', a, ' tau= ', tau)
    ocr = a*mon*np.exp(-a*mon*tau)
    ocr_meas = ocr + 2*np.random.normal(size=len(ocr), scale=30.0)
    (params,msg) = fit_deadtime(mon,ocr_meas)
    tau = params[0]
    a   = params[1]
    #print(msg)
    print( 'a_fit= ',a,' tau_fit=', tau)

    ocr = 0.3 * 1/tau
    icr = calc_icr(ocr,tau)
    print( 'max icr = ', 1/tau)
    print( 'max ocr = ', np.exp(-1)/tau)
    print( 'ocr= ', ocr, ' icr_calc= ',icr)

    rt = 1.
    lt = 1.
    cor = correction_factor(rt,lt,icr,ocr)
    print( 'cor= ', cor)




# 9/20/2010: Y.Choi.
# Reflectivity calculation using Parratt's recursive formula
# Calculate reflected intensity as a function of angle or energy
# Needs accesss to index of refraction data
# layer0 is vaccum, layer1 is top film layer ...

import math
import numpy as np
from xraydb import xray_delta_beta

class Layer:
    def __init__(self, tag, composition='Si', density=2.33, thickness=1000.1234, rms=1e-3):
        self.tag=tag                        # layer name, ex), 'top layer','film1'...
        self.composition=composition        # chemical formula ex) SiO2
        self.density=density                # nominal density in g/cm^3
        self.relden=1.0                     # relative density 1.0=100% of nominal density
        self.thickness=thickness            # in Angstroms  1e8 A = 1cm
        if rms==0:  rms=1e-9                # cannot handle 0 roughness.
        self.rms=rms                        # rms interface roughness in Agnstroms
        self.la=1.0                         # absoprtion length in cm, for inicident x-ray
        self.laFY=1.0                       # absorption length in Angstrom, for emitted fluorescence
        self.trans=0.5                      # transmission coefficient
        self.absrp=0.5                      # absorption coefficient
        self.delta=0.1                      # index of refraction, real part
        self.beta=0.1                       # index of refraction, imaginary part
        self.kz=1.0+1.0J                    # wavenumber, complex [wavenumZ]
        self.rr=0+0J                        # Fresnel coefficient, [FresnelR]
        self.rf=0+0J                        # Fresnel coefficient with roughness, refl:[FresnelR]+[RoughGauss]
        self.rrr=0+0J                       # reflectivity, [Reflect]
        self.tt=0+0J                        # Fresnel coeffcient, [FresnelT]
        self.tf=0+0J                        # Fresnel coefficient with roughness, for transmission
        self.ttt=0+0J                       # transmission, [Transmit]
        self.E_t=0+0J                       # electric field, transmission
        self.E_r=0+0J                       # electric field, relfection
        self.interface=0                    # interface depth, air/film is 0 for air layer
        self.index=-1                       # air=0, film1=1, ...substrate=n
        self.Nat=1                          # atomic number density for element of interest
    def set_DenThickRms(self, RelativeDensity, thickness, roughness):
        self.relden=RelativeDensity
        self.thickness=thickness
        self.rms=roughness

    def get_index(self, energy=8370.0, indexNat=0):   # x-ray energy in eV, relative density
        # MN replace...
        # temp=readf1f2a.get_delta(self.composition, self.density*self.relden, energy, indexNat)  # used to be fluo
        #self.delta=temp[0]
        #self.beta=temp[1]
        #self.la=temp[2]                     # in cm,
        #self.Nat=temp[3]
        delta, beta, la = xray_delta_beta(self.composition,
                                          self.density*self.relden, energy)
        self.delta, self.beta, self.la = delta, beta, la
        #la attenuation length cm, NumLayer for multiple layers
        self.trans=math.exp(-self.thickness*1e8/la)
        self.absrp=1.0-self.trans

    def cal_kz(self, k0, th_rad):   # calculate wavenumber in each layer
        temp = np.sqrt((math.sin(th_rad))**2 - 2.0*self.delta - 2.0*self.beta*1.0j)
        self.kz = k0*(temp.real + abs(temp.imag)*1.0J)
        #kz = k0 * sqrt((sin(th_rad))^2 - 2*delta - 2*beta*i)
    def cal_rr(self, kz0, kz1):
        # Fresnel coefficient, [FresnelR] input: kz_(j), kz_(j+1)
        temp = (kz0-kz1)/(kz0+kz1)
        self.rr = temp
    def cal_rf(self, rr, kz0, kz1):
        # add roughness effect to Fresnel coefficient, [RoughGauss], input:kz_(j), kz_(j+1), rr, rms
        temp = rr*np.exp(-2.0*kz0*kz1*self.rms)
        self.rf = temp
    def cal_rrr(self, rf0, rrr1, kz1, d1):  # reflectivity, [Reflect]
        v = 2.0*kz1*d1
        u = -v.imag + v.real*1.0J
        z1 = rrr1*np.exp(u) + rf0
        z2 = 1.0 + rf0*rrr1*np.exp(u)
        z=z1/z2
        self.rrr=z
    def cal_tt(self, kz0, kz1):
        #Fresnel coefficient [FresnelT], upto here 0: ThisLayer_(i), 1: BelowLayer_(i+1)
        temp = 2*kz0/(kz0+kz1)
        self.tt=temp
    def cal_ttt(self, tf2, rf2, rrr0, d0, ttt2, kz0):
        #Fresnel coefficient [FresnelT], 0:ThisLayer_(i), 2: AboveLayer_(i-1)
        v1 = -2.0*d0*kz0.imag + (2.0*d0*kz0.real)*1.0J
        v2 = np.exp(v1)
        v3 = -d0*kz0.imag + (d0*kz0.real)*1.0J
        v4 = np.exp(v3)
        v5 = tf2*(ttt2*v4)
        v6 = rf2*(rrr0*v2)
        v7 = v6.real+1.0 + (v6.imag)*1.0J
        z=v5/v7
        self.ttt=z


def reflectivity(eV0=14000.0, th_deg=[], mat=[], thick=[], den=[], rough=[], tag=[],  depths=[], xsw_mode=0):
#   --------------------set default values------------------------
    if th_deg==[]:  # set default th list
        th_min=0.0         # minimum th
        th_max=1.0          # maximum th
        th_step=0.002      # stepsize th
        th_deg = np.arange(th_min, th_max+th_step, th_step)
        th=[]               #   theta(incident angle) in radian
        qz=[]               #   momentum transfer (1/Angstrom)
    if mat==[] or thick==[] or den==[] or rough==[]:
    #set defalut layer material, air, film1, film2, substrate
        tag=['air', 'film', 'underlayer', 'substrate']
        mat=['N1.56O0.48C0.03Ar0.01Kr0.000001Xe0.0000009', 'Pt', 'Cr', 'Si']
        thick=[0., 200., 50., 10000]  # air, film1, film2, substrate
        den=[1.e-10, 21.45, 7.19, 2.33]
        rough=[1.0, 1.0, 1.0, 1.0]
    if depths==[]:  # set default depth list
        depths=[0.0]    # in Angstroms
    depth_num=len(depths)
    layers=[]
    for (ii, d) in enumerate(thick):
        layer0=Layer(tag[ii], mat[ii], den[ii])
        layer0.set_DenThickRms(1.0, thick[ii], rough[ii])
        layers.append(layer0)
    NumLayers=len(layers)       # this value is (# of film layer + 2[air, substrate])
    FilmThick=0.0  # film thickness
    for (ix, item) in enumerate(layers):
        item.get_index(eV0)         # get index of refraction, delta, beta for eV0
        FilmThick+=item.thickness   # get total film thickness
        item.interface=FilmThick    # interface location: depth from surface
        item.index=ix               # layer index
    WaveLength=12.39842/(eV0/1000.0)  # in Angstroms, 1e-8cm = 1 Angstrom
    k0=2.*math.pi/WaveLength
    for (ix, angle) in enumerate(th_deg):   # convert th_deg to th(rad) and qz(1/A)
        temp=angle/180.0*math.pi
        th.append(temp)                     # th in radian
        temp1=4.*math.pi/WaveLength*math.sin(temp)   #qz=4pi/lambda*sin(th)*2k0sin(th)
        qz.append(temp1)
    NumTh=len(th)
    out_parratt='SimpleParratt.txt'     # output file name
    fp=open(out_parratt, 'w')
    ListOut=[]
    for angle in th:
        # define wavenumber within each layer
        for layer in layers:
            layer.cal_kz(k0, angle)
        # calculate reflection and transmission coefficients
        for ix in range(NumLayers-1):
            layer=layers[ix]
            kz0=layers[ix].kz
            kz1=layers[ix+1].kz
            layer.cal_rr(kz0, kz1)                      # Fresnel coefficient, rr
            layer.cal_rf( layer.rr, kz0, kz1)           # Fresnel coefficient with roughness, rf
            layer.cal_tt(kz0, kz1)                      # Fresnel coefficient, tt, in e-field intensity
            layer.tf=layer.rf+1.0                       # T=1+R
        # calculate reflectivity
        layers[NumLayers-1].rrr=0.0                     # nothing reflects back from below the substrate
        layers[NumLayers-2].rrr=layers[NumLayers-2].rf  # interface between the substrate and layer above it
        ixs = np.arange(NumLayers-3, -1, -1)           # start from second interface from the substrate to 0
        for ix in ixs:          # calculate RRR
            BelowLayer=layers[ix+1]
            ThisLayer=layers[ix]
            ThisLayer.cal_rrr(ThisLayer.rf, BelowLayer.rrr, BelowLayer.kz, BelowLayer.thickness)
        intenR = (layers[0].rrr.real)**2 + (layers[0].rrr.imag)**2
        # intenR is reflected intensity
        # calculate transmission
        layers[0].ttt=1.0 + 0J
        ixs=np.arange(1, NumLayers-1, 1)     # from 1 to <(NumLayers-1)
        for ix in ixs:          # Calculate TTT
            ThisLayer=layers[ix]
            AboveLayer=layers[ix-1]
            ThisLayer.cal_ttt(AboveLayer.tf, AboveLayer.rf, ThisLayer.rrr, \
                              ThisLayer.thickness, AboveLayer.ttt, ThisLayer.kz)
        ixs=np.arange(0, NumLayers-1, 1)     # from 1 to <(NumLayers-1)
        for ix in ixs:
            ThisLayer=layers[ix]
            ThisLayer.E_t=ThisLayer.ttt
            ThisLayer.E_r=ThisLayer.ttt*ThisLayer.rrr
        layers[NumLayers-1].E_t = layers[NumLayers-2].tf*layers[NumLayers-2].ttt
##        Next, the current depth within the film structure is searched.
##        depth=0 is air/film, depth>0 film, depth<0 above film
##        xx=0 is film/substrate, xx<0 inside substrate, xx>0 above substrate
##        for reflectivity only, fix depth=0 and skip full xsw calculation
        layer_index=0   # air/layer1
        netFY=0.0; EFIat0=0.0; FY=0.0
        TransFY=1.0;    TransFYList=[]
        EFI_atTH=[]     # EFI as a function of depth at current th
        EFI = 0
        #print( depth_num)
        for depth in depths:
            # calculate e-field intensity at each depth
            xx=FilmThick-depth
            ThisLayer = layers[layer_index]
            if layer_index==(NumLayers-1):   #inside substrate
                cph = -xx*ThisLayer.kz
                cph_t = np.exp(cph*(0.0+1.0J))
                Etot = ThisLayer.E_t*cph_t
            elif layer_index==(NumLayers-2):    # layer above the substrate
                d_n=0
                cph = ThisLayer.kz*(xx-d_n)
                cph_t = np.exp(cph*(0.0-1.0J))
                cph_r = np.exp(cph*(0.0+1.0J))
                Etot = ThisLayer.E_t*cph_t + ThisLayer.E_r*cph_r
            else:
                d_n=FilmThick-ThisLayer.interface
                cph = ThisLayer.kz*(xx-d_n)
                cph_t = np.exp(cph*(0.0-1.0J))
                cph_r = np.exp(cph*(0.0+1.0J))
                Etot = ThisLayer.E_t*cph_t + ThisLayer.E_r*cph_r
            EFI = (Etot.real)**2 + (Etot.imag)**2
        if xsw_mode:
            pass
        temp=angle*180.0/math.pi
        out=str(temp)+'  '+str(intenR)+'  '+str(EFI)+'\n'
        ListOut.append([temp, intenR, EFI])
        fp.write(out)
    fp.close()
    #for layer in layers:
    #    print( layer.tag, layer.composition, layer.thickness, layer.rms, layer.density)
    return ListOut  #list of [[th1, refl1], [th2, refl2] ...]




if __name__=='__main__':
    #test
    # reload(readf1f2a)
    import time
    a=time.monotonic()
    reflectivity()
    print(time.monotonic()-a, 'seconds')



#!/usr/bin/env python

import cmd
import os
import sys
import numpy
import matplotlib

from .symboltable import SymbolTable
from .interpreter import Interpreter
from .site_config import history_file, show_site_config
from .version import __version__, __date__, make_banner
from .inputText import InputText
from .larchlib import StdWriter
from .utils import uname

HAS_READLINE = False
try:
    import readline
    HAS_READLINE = True
except ImportError:
    pass

HAS_WXPYTHON = False
try:
    import wx
    HAS_WXPYTHON = True
except ImportError:
    wx = None


class shell(cmd.Cmd):
    def __init__(self,  completekey='tab', debug=False, quiet=False,
                 stdin=None, stdout=None, banner_msg=None,
                 maxhist=25000, with_wx=False, with_plugins=True):

        self.debug  = debug
        cmd.Cmd.__init__(self,completekey='tab')
        homedir = os.environ.get('HOME', os.getcwd())

        if stdin is not None:
            sys.stdin = stdin
        if stdout is not None:
            sys.stdout = stdout
        self.stdin = sys.stdin
        self.stdout = sys.stdout

        if HAS_READLINE:
            try:
                readline.read_history_file(history_file)
            except IOError:
                print('could not read history from %s' % history_file)

        self.larch = Interpreter(with_plugins=with_plugins,
                                 historyfile=history_file,
                                 maxhistory=maxhist)
        self.larch.writer = StdWriter(_larch=self.larch)

        if with_wx and HAS_WXPYTHON:
            symtable = self.larch.symtable

            app = wx.App(redirect=False, clearSigInt=False)
            symtable.set_symbol('_sys.wx.wxapp', app)
            symtable.set_symbol('_sys.wx.force_wxupdate', False)
            symtable.set_symbol('_sys.wx.parent', None)

            def onCtrlC(*args, **kws): return 0

            from .wxlib import inputhook
            symtable.set_symbol('_sys.wx.inputhook', inputhook)
            symtable.set_symbol('_sys.wx.ping',   inputhook.ping)
            inputhook.ON_INTERRUPT = onCtrlC
            inputhook.WXLARCH_SYM = symtable

        self.prompt = self.larch.input.prompt
        writer = self.larch.writer
        self.color_writer = uname != 'win' and hasattr(writer, 'set_textstyle')
        if not quiet:
            if banner_msg is None:
                banner_msg = make_banner(mods=[wx])
            if self.color_writer:
                writer.set_textstyle('error')
            writer.write(banner_msg)
            writer.write("\n")
            if self.color_writer:
                writer.set_textstyle('text')

        self.larch.run_init_scripts()

    def on_exit(self, text=None):
        trim_last = False
        if text is not None:
            trim_last = text.strip() in ('quit', 'exit')
        try:
            self.larch.input.history.save(trim_last=trim_last)
        except PermissionError:
            print("Warning: could not save session history -- permission denied")
        sys.exit()

    def do_quit(self, text):
        self.on_exit(text=text)

    def do_exit(self, text):
        self.on_exit(text=text)

    def emptyline(self):
        pass

    def onecmd(self, txt):
        return self.default(txt)

    def do_help(self, txt):
        if txt.startswith('(') and txt.endswith(')'):
            txt = txt[1:-1]
        elif txt.startswith("'") and txt.endswith("'"):
            txt = txt[1:-1]
        elif txt.startswith('"') and txt.endswith('"'):
            txt = txt[1:-1]
        self.default("help(%s)" % txt)

    def do_shell(self, txt):
        os.system(txt)

    def larch_execute(self, text):
        self.default(text)

    def default(self, text):
        if text.strip() in ('quit', 'exit', 'quit()', 'exit()', 'EOF'):
            self.on_exit(text)
        ret = self.larch.eval(text, fname='<stdin>', lineno=0)
        if self.larch.error:
            self.larch.input.clear()
            if self.color_writer:
                self.larch.writer.set_textstyle('error')
            self.larch.show_errors()
            if self.color_writer:
                self.larch.writer.set_textstyle('text')
        if ret is not None:
            self.larch.writer.write("%s\n" % repr(ret))

        self.larch.writer.flush()
        self.prompt = self.larch.input.next_prompt

if __name__ == '__main__':
    t = shell(debug=True).cmdloop()



#!/usr/bin/env python

from copy import copy, deepcopy
import random
import numpy as np
from scipy.stats import f

import lmfit
from lmfit import Parameter as lmfitParameter
from lmfit import (Parameters, Minimizer, conf_interval,
                   ci_report, conf_interval2d)

from lmfit.minimizer import eval_stderr, MinimizerResult
from lmfit.model import (ModelResult, save_model, load_model,
                         save_modelresult, load_modelresult)
from lmfit.confidence import f_compare
from uncertainties import ufloat, correlated_values

from ..symboltable import Group, isgroup

def isParameter(x):
    return (isinstance(x, Parameter) or
            x.__class__.__name__ == 'Parameter')

def param_value(val):
    "get param value -- useful for 3rd party code"
    while isinstance(val, lmfitParameter):
        val = val.value
    return val

def f_test(ndata, nvars, chisquare, chisquare0, nfix=1):
    """return the F-test value for the following input values:
    f = f_test(ndata, nparams, chisquare, chisquare0, nfix=1)

    nfix = the number of fixed parameters.
    """
    return f_compare(ndata, nvars, chisquare, chisquare0, nfix=1)

def confidence_report(conf_vals, **kws):
    """return a formatted report of confidence intervals calcualted
    by confidence_intervals
    """
    return ci_report(conf_vals)


class ParameterGroup(Group):
    """
    Group for Fitting Parameters
    """
    def __init__(self, name=None, **kws):
        if name is not None:
            self.__name__ = name
        if '_larch' in kws:
            kws.pop('_larch')
        self.__params__ = Parameters()
        Group.__init__(self)
        self.__exprsave__ = {}
        for key, val in kws.items():
            expr = getattr(val, 'expr', None)
            if expr is not None:
                self.__exprsave__[key] =  expr
                val.expr = None
            setattr(self, key, val)

        for key, val in self.__exprsave__.items():
            self.__params__[key].expr = val


    def __repr__(self):
        return '<Param Group {:s}>'.format(self.__name__)

    def __setattr__(self, name, val):
        if isinstance(val, lmfitParameter):
            if val.name != name:
                # allow 'a=Parameter(2, ..)' to mean Parameter(name='a', value=2, ...)
                nval = None
                try:
                    nval = float(val.name)
                except (ValueError, TypeError):
                    pass
                if nval is not None:
                    val.value = nval
            self.__params__.add(name, value=val.value, vary=val.vary, min=val.min,
                              max=val.max, expr=val.expr, brute_step=val.brute_step)
            val = self.__params__[name]
        elif hasattr(self, '__params__') and not name.startswith('__'):
            self.__params__._asteval.symtable[name] = val
        self.__dict__[name] = val

    def __add(self, name, value=None, vary=True, min=-np.inf, max=np.inf,
              expr=None, stderr=None, correl=None, brute_step=None):
        if expr is None and isinstance(value, str):
            expr = value
            value = None
        if self.__params__  is not None:
            self.__params__.add(name, value=value, vary=vary, min=min, max=max,
                              expr=expr, brute_step=brute_step)
            self.__params__[name].stderr = stderr
            self.__params__[name].correl = correl
            self.__dict__[name] = self.__params__[name]


def param_group(**kws):
    "create a parameter group"
    return ParameterGroup(**kws)

def randstr(n):
    return ''.join([chr(random.randint(97, 122)) for i in range(n)])

class Parameter(lmfitParameter):
    """A Parameter that can be nameless"""
    def __init__(self, name=None, value=None, vary=True, min=-np.inf, max=np.inf,
                 expr=None, brute_step=None, user_data=None):
        if name is None:
            name = randstr(8)
        self.name = name
        self.user_data = user_data
        self.init_value = value
        self.min = min
        self.max = max
        self.brute_step = brute_step
        self.vary = vary
        self._expr = expr
        self._expr_ast = None
        self._expr_eval = None
        self._expr_deps = []
        self._delay_asteval = False
        self.stderr = None
        self.correl = None
        self.from_internal = lambda val: val
        self._val = value
        self._init_bounds()
        lmfitParameter.__init__(self, name, value=value, vary=vary,
                                min=min, max=max, expr=expr,
                                brute_step=brute_step,
                                user_data=user_data)

def param(*args, **kws):
    "create a fitting Parameter as a Variable"
    if len(args) > 0:
        a0 = args[0]
        if isinstance(a0, str):
            kws.update({'expr': a0})
        elif isinstance(a0, (int, float)):
            kws.update({'value': a0})
        else:
            raise ValueError("first argument to param() must be string or number")
        args = args[1:]
    if '_larch' in kws:
        kws.pop('_larch')
    if 'vary' not in kws:
        kws['vary'] = False

    return Parameter(*args, **kws)

def guess(value,  **kws):
    """create a fitting Parameter as a Variable.
    A minimum or maximum value for the variable value can be given:
       x = guess(10, min=0)
       y = guess(1.2, min=1, max=2)
    """
    kws.update({'vary':True})
    return param(value, **kws)

def is_param(obj):
    """return whether an object is a Parameter"""
    return isParameter(obj)

def group2params(paramgroup):
    """take a Group of Parameter objects (and maybe other things)
    and put them into a lmfit.Parameters, ready for use in fitting
    """
    if isinstance(paramgroup, ParameterGroup):
        return paramgroup.__params__

    params = Parameters()
    if paramgroup is not None:
        for name in dir(paramgroup):
            par = getattr(paramgroup, name)
            if isParameter(par):
                params.add(name, value=par.value, vary=par.vary,
                           min=par.min, max=par.max,
                           brute_step=par.brute_step)

            else:
                params._asteval.symtable[name] = par

        # now set any expression (that is, after all symbols are defined)
        for name in dir(paramgroup):
            par = getattr(paramgroup, name)
            if isParameter(par) and par.expr is not None:
                params[name].expr = par.expr

    return params

def params2group(params, paramgroup):
    """fill Parameter objects in paramgroup with
    values from lmfit.Parameters
    """
    for name, param in params.items():
        this = getattr(paramgroup, name, None)
        if isParameter(this):
            for attr in ('value', 'vary', 'stderr', 'min', 'max', 'expr',
                         'name', 'correl', 'brute_step', 'user_data'):
                setattr(this, attr, getattr(param, attr, None))
            if this.stderr is not None:
                try:
                    this.uvalue = ufloat(this.value, this.stderr)
                except:
                    pass


def minimize(fcn, paramgroup, method='leastsq', args=None, kws=None,
             scale_covar=True, iter_cb=None, reduce_fcn=None, nan_polcy='omit',
             **fit_kws):
    """
    wrapper around lmfit minimizer for Larch
    """
    if isinstance(paramgroup, ParameterGroup):
        params = paramgroup.__params__
    elif isgroup(paramgroup):
        params = group2params(paramgroup)
    elif isinstance(Parameters):
        params = paramgroup
    else:
        raise ValueError('minimize takes ParamterGroup or Group as first argument')

    if args is None:
        args = ()
    if kws is None:
        kws = {}

    def _residual(params):
        params2group(params, paramgroup)
        return fcn(paramgroup, *args,  **kws)

    fitter = Minimizer(_residual, params, iter_cb=iter_cb,
                       reduce_fcn=reduce_fcn, nan_policy='omit', **fit_kws)

    result = fitter.minimize(method=method)
    params2group(result.params, paramgroup)

    out = Group(name='minimize results', fitter=fitter, fit_details=result,
                chi_square=result.chisqr, chi_reduced=result.redchi)

    for attr in ('aic', 'bic', 'covar', 'params', 'nvarys',
                 'nfree', 'ndata', 'var_names', 'nfev', 'success',
                 'errorbars', 'message', 'lmdif_message', 'residual'):
        setattr(out, attr, getattr(result, attr, None))
    return out

def fit_report(fit_result, modelpars=None, show_correl=True, min_correl=0.1,
               sort_pars=True, **kws):
    """generate a report of fitting results
    wrapper around lmfit.fit_report

    The report contains the best-fit values for the parameters and their
    uncertainties and correlations.

    Parameters
    ----------
    fit_result : result from fit
       Fit Group output from fit, or lmfit.MinimizerResult returned from a fit.
    modelpars : Parameters, optional
       Known Model Parameters.
    show_correl : bool, optional
       Whether to show list of sorted correlations (default is True).
    min_correl : float, optional
       Smallest correlation in absolute value to show (default is 0.1).
    sort_pars : bool or callable, optional
       Whether to show parameter names sorted in alphanumerical order. If
       False, then the parameters will be listed in the order they
       were added to the Parameters dictionary. If callable, then this (one
       argument) function is used to extract a comparison key from each
       list element.

    Returns
    -------
    string
       Multi-line text of fit report.


    """
    result = getattr(fit_result, 'fit_details', fit_result)
    if isinstance(result,  MinimizerResult):
        return lmfit.fit_report(result, modelpars=modelpars,
                                show_correl=show_correl,
                                min_correl=min_correl, sort_pars=sort_pars)
    elif isinstance(result,  ModelResult):
        return result.fit_report(modelpars=modelpars,
                                 show_correl=show_correl,
                                 min_correl=min_correl, sort_pars=sort_pars)
    else:
        result = getattr(fit_result, 'params', fit_result)
        if isinstance(result,  Parameters):
            return lmfit.fit_report(result, modelpars=modelpars,
                                    show_correl=show_correl,
                                    min_correl=min_correl, sort_pars=sort_pars)
        else:
            try:
                result = group2params(fit_result)
                return lmfit.fit_report(result, modelpars=modelpars,
                                        show_correl=show_correl,
                                        min_correl=min_correl, sort_pars=sort_pars)
            except (ValueError, AttributeError):
                pass
    return "Cannot make fit report with %s" % repr(fit_result)


def confidence_intervals(fit_result, sigmas=(1, 2, 3), **kws):
    """calculate the confidence intervals from a fit
    for supplied sigma values

    wrapper around lmfit.conf_interval
    """
    fitter = getattr(fit_result, 'fitter', None)
    result = getattr(fit_result, 'fit_details', None)
    return conf_interval(fitter, result, sigmas=sigmas, **kws)

def chi2_map(fit_result, xname, yname, nx=11, ny=11, sigma=3, **kws):
    """generate a confidence map for any two parameters for a fit

    Arguments
    ==========
       minout   output of minimize() fit (must be run first)
       xname    name of variable parameter for x-axis
       yname    name of variable parameter for y-axis
       nx       number of steps in x [11]
       ny       number of steps in y [11]
       sigma    scale for uncertainty range [3]

    Returns
    =======
        xpts, ypts, map

    Notes
    =====
     1.  sigma sets the extent of values to explore:
              param.value +/- sigma * param.stderr
    """
    #
    fitter = getattr(fit_result, 'fitter', None)
    result = getattr(fit_result, 'fit_details', None)
    if fitter is None or result is None:
        raise ValueError("chi2_map needs valid fit result as first argument")

    c2_scale = fit_result.chi_square / result.chisqr

    def scaled_chisqr(ndata, nparas, new_chi, best_chi, nfix=1.):
        """return scaled chi-sqaure, instead of probability"""
        return new_chi * c2_scale

    x = result.params[xname]
    y = result.params[yname]
    xrange = (x.value + sigma * x.stderr, x.value - sigma * x.stderr)
    yrange = (y.value + sigma * y.stderr, y.value - sigma * y.stderr)

    return conf_interval2d(fitter, result, xname, yname,
                           limits=(xrange, yrange),
                           prob_func=scaled_chisqr,
                           nx=nx, ny=ny, **kws)

def _Parameters(*arg, **kws):
    return Parameters(*arg, **kws)


_larch_name = '_math'
exports = {'param': param,
           'guess': guess,
           'param_group': param_group,
           'confidence_intervals': confidence_intervals,
           'confidence_report': confidence_report,
           'f_test': f_test, 'chi2_map': chi2_map,
           'is_param': isParameter,
           'isparam': isParameter,
           'minimize': minimize,
           'ufloat': ufloat,
           'fit_report': fit_report,
           'Parameters': _Parameters,
           'Parameter': Parameter,
           'lm_minimize': minimize,
           'lm_save_model': save_model,
           'lm_load_model': load_model,
           'lm_save_modelresult': save_modelresult,
           'lm_load_modelresult': load_modelresult,
           }

for name in ('BreitWignerModel', 'ComplexConstantModel',
             'ConstantModel', 'DampedHarmonicOscillatorModel',
             'DampedOscillatorModel', 'DonaichModel',
             'ExponentialGaussianModel', 'ExponentialModel',
             'ExpressionModel', 'GaussianModel', 'Interpreter',
             'LinearModel', 'LognormalModel', 'LorentzianModel',
             'MoffatModel', 'ParabolicModel', 'Pearson7Model',
             'PolynomialModel', 'PowerLawModel',
             'PseudoVoigtModel', 'QuadraticModel',
             'RectangleModel', 'SkewedGaussianModel',
             'StepModel', 'StudentsTModel', 'VoigtModel'):
    val = getattr(lmfit.models, name, None)
    if val is not None:
        exports[name] = val

_larch_builtins = {'_math': exports}



__DOC__ = '''
XAFS Functions for Larch, essentially Ifeffit 2

The functions here include (but are not limited to):

function         description
------------     ------------------------------
pre_edge         pre_edge subtraction, normalization
autobk           XAFS background subtraction (mu(E) to chi(k))
xftf             forward XAFS Fourier transform (k -> R)
xftr             backward XAFS Fourier transform, Filter (R -> q)
ftwindow         create XAFS Fourier transform window

feffpath         create a Feff Path from a feffNNNN.dat file
path2chi         convert a single Feff Path to chi(k)
ff2chi           sum a set of Feff Paths to chi(k)

feffit_dataset   create a Dataset for Feffit
feffit_transform create a Feffit transform group
feffit           fit a set of Feff Paths to Feffit Datasets
feffit_report    create a report from feffit() results
'''


from .xafsutils import KTOE, ETOK, set_xafsGroup, etok, ktoe, guess_energy_units
from .xafsft import xftf, xftr, xftf_fast, xftr_fast, ftwindow, xftf_prep
from .pre_edge import pre_edge, preedge, find_e0, pre_edge_baseline, prepeaks_setup
from .feffdat import FeffDatFile, FeffPathGroup, feffpath, path2chi, ff2chi
from .feffit import (FeffitDataSet, TransformGroup, feffit,
                     feffit_dataset, feffit_transform, feffit_report)

from .autobk import autobk
from .mback import mback, mback_norm
from .diffkk import diffkk, diffKKGroup
from .fluo import fluo_corr

from .feffrunner import FeffRunner, feffrunner, feff6l, feff8l, find_exe
from .feff8lpath import feff8_xafs



from .cauchy_wavelet import cauchy_wavelet
from .deconvolve import xas_convolve, xas_deconvolve
from .estimate_noise import estimate_noise
from .rebin_xafs import rebin_xafs, sort_xafs
from .sigma2_models import sigma2_eins, sigma2_debye, sigma2_correldebye


def _larch_init(_larch):
    """initialize xafs"""
    # initialize _xafs._feff_executable
    feff6_exe = find_exe('feff6l')
    _larch.symtable.set_symbol('_xafs._feff_executable', feff6_exe)
    

_larch_groups = (diffKKGroup, FeffRunner, FeffDatFile, FeffPathGroup,
                 TransformGroup, FeffitDataSet)

_larch_builtins = {'_xafs': dict(autobk=autobk, etok=etok, ktoe=ktoe,
                                 guess_energy_units=guess_energy_units,
                                 diffkk=diffkk, xftf=xftf, xftr=xftr,
                                 xftf_prep=xftf_prep, xftf_fast=xftf_fast,
                                 xftr_fast=xftr_fast, ftwindow=ftwindow,
                                 find_e0=find_e0, pre_edge=pre_edge,
                                 prepeaks_setup=prepeaks_setup,
                                 pre_edge_baseline=pre_edge_baseline,
                                 mback=mback, mback_norm=mback_norm,
                                 cauchy_wavelet=cauchy_wavelet,
                                 xas_deconvolve=xas_deconvolve,
                                 xas_convolve=xas_convolve,
                                 fluo_corr=fluo_corr,
                                 estimate_noise=estimate_noise,
                                 rebin_xafs=rebin_xafs,
                                 sort_xafs=sort_xafs,
                                 sigma2_eins=sigma2_eins,
                                 sigma2_debye=sigma2_debye, feffit=feffit,
                                 feffit_dataset=feffit_dataset,
                                 feffit_transform=feffit_transform,
                                 feffit_report=feffit_report,
                                 feffrunner=feffrunner, feff6l=feff6l,
                                 feff8l=feff8l, feffpath= feffpath,
                                 path2chi=path2chi, ff2chi=ff2chi,
                                 feff8_xafs=feff8_xafs)}



#!/usr/bin/python
import sys
import os
import time
from configparser import ConfigParser

from epics import Device, caget, caput, poll
from epics.devices.mca import MCA, ROI, OrderedDict
from epics.devices.ad_mca import ADMCA, ADMCAROI

MAX_ROIS = 32

class ADFileMixin(object):
    """mixin class for Xspress3"""
    def filePut(self, attr, value, **kws):
        return self.put("%s%s" % (self.filesaver, attr), value, **kws)

    def fileGet(self, attr, **kws):
        return self.get("%s%s" % (self.filesaver, attr), **kws)

    def setFilePath(self, pathname):
        fullpath = os.path.join(self.fileroot, pathname)
        return self.filePut('FilePath', fullpath)

    def setFileTemplate(self,fmt):
        return self.filePut('FileTemplate', fmt)

    def setFileWriteMode(self,mode):
        return self.filePut('FileWriteMode', mode)

    def setFileName(self,fname):
        return self.filePut('FileName', fname)

    def nextFileNumber(self):
        self.setFileNumber(1+self.fileGet('FileNumber'))

    def setFileNumber(self, fnum=None):
        if fnum is None:
            self.filePut('AutoIncrement', 1)
        else:
            self.filePut('AutoIncrement', 0)
            return self.filePut('FileNumber',fnum)

    def getLastFileName(self):
        return self.fileGet('FullFileName_RBV',as_string=True)

    def FileCaptureOn(self):
        return self.filePut('Capture', 1)

    def FileCaptureOff(self):
        return self.filePut('Capture', 0)

    def setFileNumCapture(self,n):
        return self.filePut('NumCapture', n)

    def FileWriteComplete(self):
        return (0==self.fileGet('WriteFile_RBV') )

    def getFileTemplate(self):
        return self.fileGet('FileTemplate_RBV',as_string=True)

    def getFileName(self):
        return self.fileGet('FileName_RBV',as_string=True)

    def getFileNumber(self):
        return self.fileGet('FileNumber_RBV')

    def getFilePath(self):
        return self.fileGet('FilePath_RBV',as_string=True)

    def getFileNameByIndex(self,index):
        return self.getFileTemplate() % (self.getFilePath(), self.getFileName(), index)

class Xspress3BaseMixin(object):
    """xspress3 mixin -- triggers, acquire, etc"""
    def useExternalTrigger(self):
        self.TriggerMode = 3

    def useInternalTrigger(self):
        self.TriggerMode = 1

    def setTriggerMode(self, mode):
        self.TriggerMode = mode

    def start(self, capture=True):
        time.sleep(.05)
        if capture:
            self.FileCaptureOn()
        self.Acquire = 1

    def stop(self):
        self.Acquire = 0
        self.FileCaptureOff()

    def get_rois(self):
        return [m.get_rois() for m in self.mcas]

class Xspress3(Device, ADFileMixin, Xspress3BaseMixin):
    """Epics Xspress3.20 interface (with areaDetector2)"""

    det_attrs = ('NumImages', 'NumImages_RBV', 'Acquire', 'Acquire_RBV',
                 'ArrayCounter_RBV', 'ERASE', 'UPDATE', 'AcquireTime',
                 'TriggerMode', 'StatusMessage_RBV', 'DetectorState_RBV')

    _nonpvs = ('_prefix', '_pvs', '_delim', 'filesaver', 'fileroot',
                'pathattrs', '_nonpvs',  'nmca', 'mcas')

    pathattrs = ('FilePath', 'FileTemplate', 'FileName', 'FileNumber',
                 'Capture', 'NumCapture')

    def __init__(self, prefix, nmca=4, filesaver='HDF1:',
                 fileroot='/home/xspress3/cars5/Data'):
        if not prefix.endswith(':'):
            prefix = "%s:" % prefix
        self.nmca = nmca
        attrs = []
        attrs.extend(['%s%s' % (filesaver,p) for p in self.pathattrs])

        self.filesaver = filesaver
        self.fileroot = fileroot
        self._prefix = prefix
        self.mcas = []
        for i in range(nmca):
            imca = i+1
            dprefix = "%sdet1:" % prefix
            rprefix = "%sMCA%iROI" % (prefix, imca)
            data_pv = "%sMCA%i:ArrayData" % (prefix, imca)
            mca = ADMCA(dprefix, data_pv=data_pv, roi_prefix=rprefix)
            self.mcas.append(mca)

        Device.__init__(self, prefix, attrs=attrs, delim='')
        for attr in self.det_attrs:
            self.add_pv("%sdet1:%s" % (prefix, attr), attr)
        for i in range(nmca):
            imca = i+1
            for j in range(8):
                isca = j+1
                attr="C%iSCA%i"% (imca, isca)
                self.add_pv("%s%s:Value_RBV" % (prefix, attr), attr)
            for attr in ('TSNumPoints', 'TSControl'):
                self.add_pv("%sMCA%iROI:%s" % (prefix, imca, attr),
                            "MCA%i%s" % (imca, attr))
                self.add_pv("%sC%iSCA:%s" % (prefix, imca, attr),
                            "SCA%i%s" % (imca, attr))
        time.sleep(0.05)

    def TimeSeriesCaptureOn(self, npts=None):
        """ turns on a Time Series Capture"""
        for imca in range(len(self.mcas)):
            if npts is not None:
                self._pvs["MCA%iTSNumPoints" % (imca+1)].put(npts)
                self._pvs["SCA%iTSNumPoints" % (imca+1)].put(npts)
        time.sleep(0.025)
        for imca in range(len(self.mcas)):
            self._pvs["MCA%iTSControl" % (imca+1)].put(0)
            self._pvs["SCA%iTSControl" % (imca+1)].put(0)

    def TimeSeriesCaptureOff(self):
        """ turns off a Time Series Capture"""
        for imca in range(len(self.mcas)):
            self._pvs["MCA%iTSControl" % (imca+1)].put(2)
            self._pvs["SCA%iTSControl" % (imca+1)].put(2)

    def roi_calib_info(self):
        buff = ['[rois]']
        add = buff.append
        rois = self.mcas[0].get_rois()
        for iroi, roi in enumerate(rois):
            name = roi.Name
            hi   = roi.MinX + roi.SizeX
            if len(name.strip()) > 0 and hi > 0:
                dbuff = []
                for m in range(self.nmca):
                    dbuff.extend([roi.MinX, roi.MinX+roi.SizeX])
                dbuff = ' '.join([str(i) for i in dbuff])
                add("ROI%2.2i = %s | %s" % (iroi, name, dbuff))

        add('[calibration]')
        add("OFFSET = %s " % (' '.join(["0.000 "] * self.nmca)))
        add("SLOPE  = %s " % (' '.join(["0.010 "] * self.nmca)))
        add("QUAD   = %s " % (' '.join(["0.000 "] * self.nmca)))
        add('[dxp]')
        return buff

    def restore_rois(self, roifile):
        """restore ROI setting from ROI.dat file"""
        cp =  ConfigParser()
        cp.read(roifile)
        roidat = []
        iroi = 0
        for a in cp.options('rois'):
            if a.lower().startswith('roi'):
                name, dat = cp.get('rois', a).split('|')
                lims = [int(i) for i in dat.split()]
                lo, hi = lims[0], lims[1]
                roidat.append((name.strip(), lo, hi))

        for mca in self.mcas:
            mca.set_rois(roidat)

class Xspress310(Device, ADFileMixin, Xspress3BaseMixin):
    """Epics Xspress3.10 interface (older version)"""
    attrs = ('NumImages', 'NumImages_RBV',
             'Acquire', 'Acquire_RBV',
             'ArrayCounter_RBV',
             'ERASE', 'UPDATE', 'AcquireTime',
             'TriggerMode', 'StatusMessage_RBV',
             'DetectorState_RBV')

    _nonpvs  = ('_prefix', '_pvs', '_delim', 'filesaver',
                'fileroot', 'pathattrs', '_nonpvs', '_save_rois',
                'nmca', 'dxps', 'mcas')

    pathattrs = ('FilePath', 'FileTemplate',
                 'FileName', 'FileNumber',
                 'Capture',  'NumCapture')

    def __init__(self, prefix, nmca=4, filesaver='HDF5:',
                 fileroot='/home/xspress3/cars5/Data'):
        self.nmca = nmca
        attrs = list(self.attrs)
        attrs.extend(['%s%s' % (filesaver,p) for p in self.pathattrs])

        self.filesaver = filesaver
        self.fileroot = fileroot
        self._prefix = prefix
        self._save_rois = []
        self.mcas = [MCA(prefix, mca=i+1) for i in range(nmca)]

        Device.__init__(self, prefix, attrs=attrs, delim='')
        time.sleep(0.1)


    def select_rois_to_save(self, roilist):
        """copy rois from MCA record to arrays to be saved
        by XSPress3"""
        roilist = list(roilist)
        if len(roilist) < 4: roilist.append((50, 4050))
        pref = self._prefix
        self._save_rois = []
        for iroi, roiname in enumerate(roilist):
            label = roiname
            if isinstance(roiname, tuple):
                lo, hi = roiname
                label = '[%i:%i]'  % (lo, hi)
            else:
                rname = roiname.lower().strip()
                lo, hi = 50, 4050
                for ix in range(MAX_ROIS):
                    nm = caget('%smca1.R%iNM' % (pref, ix))
                    if nm.lower().strip() == rname:
                        lo = caget('%smca1.R%iLO' % (pref, ix))
                        hi = caget('%smca1.R%iHI' % (pref, ix))
                        break
            self._save_rois.append(label)
            for imca in range(1, self.nmca+1):
                pv_lo = "%sC%i_MCA_ROI%i_LLM" % (pref, imca, iroi+1)
                pv_hi = "%sC%i_MCA_ROI%i_HLM" % (pref, imca, iroi+1)
                caput(pv_hi, hi)
                caput(pv_lo, lo)

    def roi_calib_info(self):
        buff = ['[rois]']
        add = buff.append
        rois = self.get_rois()
        for iroi in range(len(rois[0])):
            name = rois[0][iroi].NM
            hi   = rois[0][iroi].HI
            if len(name.strip()) > 0 and hi > 0:
                dbuff = []
                for m in range(self.nmca):
                    dbuff.extend([rois[m][iroi].LO, rois[m][iroi].HI])
                dbuff = ' '.join([str(i) for i in dbuff])
                add("ROI%2.2i = %s | %s" % (iroi, name, dbuff))

        add('[calibration]')
        add("OFFSET = %s " % (' '.join(["0.000 "] * self.nmca)))
        add("SLOPE  = %s " % (' '.join(["0.010 "] * self.nmca)))
        add("QUAD   = %s " % (' '.join(["0.000 "] * self.nmca)))
        add('[dxp]')
        return buff

    def restore_rois(self, roifile):
        """restore ROI setting from ROI.dat file"""
        cp =  ConfigParser()
        cp.read(roifile)
        rois = []
        self.mcas[0].clear_rois()
        prefix = self.mcas[0]._prefix
        if prefix.endswith('.'):
            prefix = prefix[:-1]
        iroi = 0
        for a in cp.options('rois'):
            if a.lower().startswith('roi'):
                name, dat = cp.get('rois', a).split('|')
                lims = [int(i) for i in dat.split()]
                lo, hi = lims[0], lims[1]
                # print('ROI ', name, lo, hi)
                roi = ROI(prefix=prefix, roi=iroi)
                roi.LO = lo
                roi.HI = hi
                roi.NM = name.strip()
                rois.append(roi)
                iroi += 1

        poll(0.050, 1.0)
        self.mcas[0].set_rois(rois)
        cal0 = self.mcas[0].get_calib()
        for mca in self.mcas[1:]:
            mca.set_rois(rois, calib=cal0)
